#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import os
import sys

from mutagen.mp3 import HeaderNotFoundError

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.track_tags import SeratoTag
from serato_tools.utils import logger


class TrackWaveform(SeratoTag):
    GEOB_KEY = "Serato Overview"
    VERSION = (0x01, 0x05)

    def __init__(self, file_or_data: SeratoTag.FileOrData):
        super().__init__(file_or_data)

        if self.raw_data is None:
            raise ValueError("no waveform yet set")

        self.data = self._parse(self.raw_data)

    def _parse(self, data: bytes):
        fp = io.BytesIO(data)
        self._check_version(fp.read(self.VERSION_LEN))

        for x in iter(lambda: fp.read(16), b""):
            assert len(x) == 16
            yield bytearray(x)

    def draw_image(self):
        try:
            from PIL import Image, ImageColor
        except:
            logger.error('must install package "pillow"')
            raise

        img = Image.new("RGB", (240, 16), "black")
        pixels = img.load()

        for i in range(img.size[0]):
            rowdata = next(self.data)
            factor = len([x for x in rowdata if x < 0x80]) / len(rowdata)

            for j, value in enumerate(rowdata):
                # The algorithm to derive the colors from the data has no real mathematical background and was found by experimenting with different values.
                color = "hsl({hue:.2f}, {saturation:d}%, {luminance:.2f}%)".format(
                    hue=(factor * 1.5 * 360) % 360,
                    saturation=40,
                    luminance=(value / 0xFF) * 100,
                )
                pixels[i, j] = ImageColor.getrgb(color)  # pyright: ignore[reportOptionalSubscript]

        return img


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()

    try:
        tags = TrackWaveform(args.file)
    except HeaderNotFoundError:
        with open(args.file, mode="rb") as fp:
            data = fp.read()
        tags = TrackWaveform(data)

    img = tags.draw_image()
    img.show()



import os
import sys
from typing import cast
from enum import StrEnum, IntEnum

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.crate_base import CrateBase
from serato_tools.utils import get_enum_key_from_value, parse_cli_keys_and_values, SERATO_DIR, DataTypeError


class SmartCrate(CrateBase):
    TESTED_VERSIONS = ["1.0/Serato ScratchLive Smart Crate"]
    EXTENSION = ".scrate"
    DIR = "SmartCrates"
    DIR_PATH = os.path.join(SERATO_DIR, DIR)

    class RuleField(IntEnum):
        ADDED = 25
        ALBUM = 8
        ARTIST = 7
        BPM = 15
        COMMENT = 17
        COMPOSER = 22
        FILENAME = 4
        GENRE = 9
        GROUPING = 19
        KEY = 51
        LABEL = 21
        PLAYS = 79
        REMIXER = 20
        SONG = 6
        YEAR = 23

    class RuleComparison(StrEnum):
        STR_CONTAINS = "cond_con_str"
        STR_DOES_NOT_CONTAIN = "cond_dnc_str"
        STR_IS = "cond_is_str"
        STR_IS_NOT = "cond_isn_str"
        STR_DATE_BEFORE = "cond_bef_str"
        STR_DATE_AFTER = "cond_aft_str"
        TIME_IS_BEFORE = "cond_bef_time"
        TIME_IS_AFTER = "cond_aft_time"
        INT_IS_GE = "cond_greq_uint"
        INT_IS_LE = "cond_lseq_uint"

    DEFAULT_ENTRIES = [
        (CrateBase.Fields.VERSION, "1.0/Serato ScratchLive Smart Crate"),
        (CrateBase.Fields.SMARTCRATE_MATCH_ALL, [("brut", False)]),
        (CrateBase.Fields.SMARTCRATE_LIVE_UPDATE, [("brut", False)]),
        (CrateBase.Fields.SORTING, [(CrateBase.Fields.COLUMN_NAME, "key"), (CrateBase.Fields.REVERSE_ORDER, False)]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "song"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "playCount"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "artist"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "bpm"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "key"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "album"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "length"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "comment"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "added"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
    ]

    @staticmethod
    def _get_rule_field_from_val(value: int) -> str:
        return get_enum_key_from_value(value, SmartCrate.RuleField)

    @staticmethod
    def _get_rule_comparison_from_val(value: str) -> str:
        return get_enum_key_from_value(value, SmartCrate.RuleComparison)

    @staticmethod
    def _get_rule_field_from_key(key: str):
        try:
            return SmartCrate.RuleField[key.upper()]
        except KeyError as exc:
            raise KeyError(
                f"bad field {key.upper()} , must be one of {[c.name for c in SmartCrate.RuleField]}"
            ) from exc

    @staticmethod
    def _get_rule_comparison_from_key(key: str):
        try:
            return SmartCrate.RuleComparison[key.upper()]
        except KeyError as exc:
            raise KeyError(
                f"bad comparison {key.upper()} , must be one of {[c.name for c in SmartCrate.RuleComparison]}"
            ) from exc

    @staticmethod
    def _get_rule_value_type(value: str | int):
        if isinstance(value, int):
            return SmartCrate.Fields.RULE_VALUE_INTEGER
        elif isinstance(value, str):
            return SmartCrate.Fields.RULE_VALUE_TEXT
        else:
            raise TypeError(f"Bad type: {type(value)} (value: {value})")

    def _stringify_entry(self, entry: CrateBase.EntryFull, indent: int = 0) -> str:
        field, fieldname, value = entry  # pylint: disable=unused-variable
        if isinstance(value, list):
            return self._stringify_entries(value, indent)

        ret_val = str(value)
        if field == CrateBase.Fields.RULE_FIELD:
            if not isinstance(value, int):
                raise DataTypeError(value, int, field)
            ret_val += f" ({self._get_rule_field_from_val(value).lower()})"
        elif field == CrateBase.Fields.RULE_COMPARISON:
            if not isinstance(value, str):
                raise DataTypeError(value, str, field)
            ret_val += f" ({self._get_rule_comparison_from_val(value)})"

        return ret_val

    class Rule(CrateBase.EntryListCls):
        def __init__(self, entries: "CrateBase.EntryList"):
            super().__init__(entries)

            # TODO: check and ensure types

            self.comparison = self.get_value(SmartCrate.Fields.RULE_COMPARISON)
            self.field = self.get_value(SmartCrate.Fields.RULE_FIELD)

            def try_to_get_value(field: str):
                try:
                    return self.get_value(field)
                except AttributeError:
                    return None

            self.value = cast(
                str | int | float,  # float for date or no?
                (
                    try_to_get_value(SmartCrate.Fields.RULE_VALUE_INTEGER)
                    or try_to_get_value(SmartCrate.Fields.RULE_VALUE_TEXT)
                    or try_to_get_value(SmartCrate.Fields.RULE_VALUE_DATE)
                ),
            )

        def set_value(  # pylint: disable=arguments-differ # pyright: ignore[reportIncompatibleMethodOverride]
            self, value: str | int
        ):
            field = SmartCrate._get_rule_value_type(value)  # pylint: disable=protected-access
            super().set_value(field, value)

        def set_field(self, value: "SmartCrate.RuleField"):
            super().set_value(SmartCrate.Fields.RULE_FIELD, value.value)

        def set_comparison(self, value: "SmartCrate.RuleComparison"):
            super().set_value(SmartCrate.Fields.RULE_COMPARISON, value.value)

    def set_rule(self, field: "SmartCrate.RuleField", comparison: "SmartCrate.RuleComparison", value: str | int):
        rule_field_id = field.value
        rule_exists: bool = False
        new_entries: SmartCrate.EntryList = []
        for f, v in self.entries:  # pylint: disable=access-member-before-definition
            if f == CrateBase.Fields.SMARTCRATE_RULE:
                if not isinstance(v, list):
                    raise DataTypeError(v, list, f)
                rule = SmartCrate.Rule(v)
                if rule.field == rule_field_id:
                    rule.set_value(value)
                    rule.set_comparison(comparison)
                    rule_exists = True
                v = rule.to_entries()
            new_entries.append((f, v))

        if not rule_exists:
            new_rule = SmartCrate.Rule(
                [
                    (SmartCrate.Fields.RULE_COMPARISON.value, comparison.value),
                    (SmartCrate.Fields.RULE_FIELD.value, rule_field_id),
                    (SmartCrate._get_rule_value_type(value).value, value),
                ]
            )
            new_entries.append((CrateBase.Fields.SMARTCRATE_RULE.value, new_rule.to_entries()))

        self.entries = new_entries  # pylint: disable=attribute-defined-outside-init
        self._dump()

    def delete_rule(self, field: "SmartCrate.RuleField"):
        rule_field_id = field.value
        new_entries: SmartCrate.EntryList = []
        for f, v in self.entries:  # pylint: disable=access-member-before-definition
            if f == CrateBase.Fields.SMARTCRATE_RULE:
                if not isinstance(v, list):
                    raise DataTypeError(v, list, f)
                rule = SmartCrate.Rule(v)
                if rule.field == rule_field_id:
                    continue
            new_entries.append((f, v))

        self.entries = new_entries  # pylint: disable=attribute-defined-outside-init
        self._dump()

    def find_missing(self):
        raise Exception("cannot be done on smart crate")


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file_or_dir", nargs="?", help="Set the crate file, or a directory to do all files in it")
    parser.add_argument("--all", action="store_true", help="Perform actions on all crates.")
    parser.add_argument("-l", "--list_tracks", action="store_true", help="Only list tracks")
    parser.add_argument("-f", "--filenames_only", action="store_true", help="Only list track filenames")
    parser.add_argument(
        "--set_rules",
        nargs=argparse.REMAINDER,
        help="Set rules for all crates using key-value pairs like --grouping str_contains NEW --title str_does_not_contain NEW_TITLE\nCan also do --key DELETE",
    )
    args = parser.parse_args()

    if args.all:
        args.file_or_dir = SmartCrate.DIR_PATH

    if not args.file_or_dir:
        print(f"must pass a file or dir, or --all!\nfiles in {SmartCrate.DIR_PATH}:")
        print("\n".join(SmartCrate.get_crate_files()))
        sys.exit()

    set_rules = parse_cli_keys_and_values(args.set_rules) if args.set_rules else {}

    for crate_path in SmartCrate.get_crate_files(args.file_or_dir):
        crate = SmartCrate(crate_path)

        if args.set_rules:
            for key, value in set_rules.items():
                field = SmartCrate._get_rule_field_from_key(key)  # pylint: disable=protected-access
                if str(value[0]).upper() == "DELETE":
                    crate.delete_rule(field)
                else:
                    assert len(value) == 2, "must specify 2 values: a comparison and a value"
                    comparison = SmartCrate._get_rule_comparison_from_key(  # pylint: disable=protected-access
                        str(value[0])
                    )
                    crate.set_rule(field, comparison, value[1])
            crate.save()
            continue

        if args.list_tracks or args.filenames_only:
            crate.print_track_paths(filenames_only=args.filenames_only)
        else:
            print(crate)


if __name__ == "__main__":
    main()



import struct
import io

from mutagen.mp3 import MP3
from mutagen.id3 import ID3FileType, ID3
from mutagen.aiff import AIFF
from mutagen.id3._frames import GEOB

from serato_tools.utils import logger


class SeratoTrack:
    type Tagfile = ID3FileType | ID3 | AIFF
    type FileArg = str | Tagfile

    def __init__(self, file: FileArg):
        self.tagfile: SeratoTrack.Tagfile

        if isinstance(file, str):
            try:
                if file.lower().endswith(".mp3"):
                    self.tagfile = MP3(file)
                elif file.lower().endswith(".aiff"):
                    self.tagfile = AIFF(file)
            except:
                logger.error(f"Mutagen error for file {file}")
                raise
        else:
            self.tagfile = file

    def _get_geob(self, geob_key: str) -> bytes | None:
        geob_key = f"GEOB:{geob_key}"
        try:
            return self.tagfile[geob_key].data
        except KeyError:
            logger.debug(f'File is missing "{geob_key}" tag, not yet set: {self.tagfile.filename}')
            return None

    def _tag_geob(self, geob_key: str, data: bytes):
        self.tagfile[f"GEOB:{geob_key}"] = GEOB(
            encoding=0,
            mime="application/octet-stream",
            desc=geob_key,
            data=data,
        )

    def _del_tag(self, key: str):
        """
        Returns True if was deleted, False if no change
        """
        if key in self.tagfile:
            del self.tagfile[key]
            return True
        return False

    def _del_geob(self, geob_key: str):
        """
        Returns True if was deleted, False if no change
        """
        return self._del_tag(f"GEOB:{geob_key}")

    VERSION_FORMAT = "BB"
    VERSION_LEN = struct.calcsize(VERSION_FORMAT)
    type Version = tuple[int, int]

    @staticmethod
    def _check_version(given: bytes, expected: Version):
        given_version: SeratoTrack.Version = struct.unpack(SeratoTrack.VERSION_FORMAT, given)
        if given_version != expected:
            raise ValueError(f"""
                ERROR: Untested Serato tag version: {str(given_version)}
                Please contact the developer so we can get it tested and supported!
                We will have you send the file you are trying to parse so we can add support and tests for it.
                We do not want to risk damaging users' library database or crate files!

                Expected version: {str(expected)}
                """)

    @staticmethod
    def _pack_version(version: Version):
        return struct.pack(SeratoTrack.VERSION_FORMAT, *version)

    @staticmethod
    def _readbytes_gen(fp: io.BytesIO):
        for x in iter(lambda: fp.read(1), b""):
            if x == b"\00":
                break
            yield x

    @staticmethod
    def _readbytes(fp: io.BytesIO):
        return b"".join(SeratoTrack._readbytes_gen(fp))


class SeratoTag(SeratoTrack):
    GEOB_KEY: str
    VERSION: SeratoTrack.Version

    type FileOrData = SeratoTrack.FileArg | bytes

    def __init__(self, file_or_data: FileOrData):
        self.tagfile: SeratoTrack.Tagfile | None = None  # pyright: ignore[reportIncompatibleVariableOverride]
        self.raw_data: bytes | None = None

        if isinstance(file_or_data, (bytes, bytearray, memoryview)):
            self.raw_data = file_or_data
        else:
            super().__init__(file=file_or_data)
            self.raw_data = self._get_geob()

    def __repr__(self) -> str:
        return str(self.raw_data)

    def _get_geob(self):  # pylint: disable=arguments-differ  # pyright: ignore[reportIncompatibleMethodOverride]
        return super()._get_geob(self.GEOB_KEY)

    def _tag_geob(self):  # pylint: disable=arguments-differ  # pyright: ignore[reportIncompatibleMethodOverride]
        if self.raw_data is None:
            return
        return super()._tag_geob(self.GEOB_KEY, self.raw_data)

    def _del_geob(self):  # pylint: disable=arguments-differ  # pyright: ignore[reportIncompatibleMethodOverride]
        return super()._del_geob(self.GEOB_KEY)

    def _check_version(  # pylint: disable=arguments-renamed  # pyright: ignore[reportIncompatibleMethodOverride]
        self, given: bytes
    ):
        return super()._check_version(given, self.VERSION)

    def _pack_version(self):  # pylint: disable=arguments-renamed  # pyright: ignore[reportIncompatibleMethodOverride]
        return super()._pack_version(self.VERSION)

    def delete(self):
        return self._del_geob()

    def save(self):
        if not self.tagfile:
            raise Exception("no tagfile, no saving")
        if (self._get_geob() != None) and (self.raw_data is None):
            raise ValueError("no data to save")
        if self.raw_data:
            self._tag_geob()
        self.tagfile.save()



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.track_cues_v1 import TrackCuesV1


class TestCase(unittest.TestCase):
    def setUp(self):
        with open(os.path.abspath("test/data/track_cues_v1.bin"), mode="rb") as fp:
            self.file_data = fp.read()

    def test_parse(self):
        tags = TrackCuesV1(self.file_data)
        self.assertEqual(tags.raw_data, self.file_data, "raw_data read")

    @unittest.skip("doesn't work with old data")
    def test_parse_and_dump(self):
        tags = TrackCuesV1(self.file_data)
        self.assertEqual(tags.raw_data, self.file_data, "raw_data read")
        tags._dump()
        self.assertEqual(tags.raw_data, self.file_data, "dump")



# Built off of https://github.com/heyitsmass/audio/blob/master/audio/beat_grid.py

import sys
from dataclasses import asdict, dataclass
from typing import List, Optional

import librosa
import numpy as np


@dataclass
class BeatGridInfo:
    bpm: float
    beat_positions: List[float]
    confidence: float
    phase_offset: float
    grid_consistency: float
    downbeats: List[float]


def analyze_beatgrid(file: str, bpm_helper: Optional[int | float] = None):
    audio_data, sample_rate = librosa.load(
        file,
        sr=None,  # sample_rate = default
    )

    # Get tempo and beat frames
    tempo, beat_times = librosa.beat.beat_track(
        y=audio_data,
        sr=sample_rate,
        trim=False,
        tightness=100,
        units="time",
        bpm=bpm_helper,
    )

    # --- Calculate grid strength ---
    # Calculate inter-beat intervals
    ibis = np.diff(beat_times)

    # Calculate variance in intervals
    variance = np.var(ibis)
    mean_ibi = np.mean(ibis)

    # Calculate normalized strength metric
    grid_consistency = float(1.0 / (1.0 + variance / mean_ibi))

    # Calculate confidence based on grid strength and tempo stability
    confidence = float(grid_consistency * (1.0 - abs(tempo - 120) / 120))

    # --- Get downbeats ---
    # Find phase offset (time to first beat)
    phase_offset = beat_times[0] if len(beat_times) > 0 else 0.0

    # Calculate onset strength
    onset_env = librosa.onset.onset_strength(y=audio_data, sr=sample_rate)

    # Get onset peaks near beat positions
    downbeats: list[float] = []
    for beat_time in beat_times[::4]:  # Check every 4th beat
        beat_frame = int(beat_time * sample_rate / 512)  # Convert to onset frames
        if beat_frame < len(onset_env):
            local_max = np.argmax(onset_env[max(0, beat_frame - 2) : min(len(onset_env), beat_frame + 3)])
            downbeats.append(float(beat_time + (local_max - 2) * 512 / sample_rate))

    return BeatGridInfo(
        bpm=float(tempo),
        beat_positions=list(beat_times),
        confidence=confidence,
        phase_offset=float(phase_offset),
        grid_consistency=grid_consistency,
        downbeats=downbeats,
    )


if __name__ == "__main__":
    from pprint import pprint

    print("Analyzing beatgrid:")
    beat_info = analyze_beatgrid(sys.argv[1])

    pprint(asdict(beat_info))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import enum
import io
import os
import struct
import sys

from mutagen.mp3 import HeaderNotFoundError

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.track_tags import SeratoTag


class TrackCuesV1(SeratoTag):
    GEOB_KEY = "Serato Markers_"
    VERSION = (0x02, 0x05)

    class EntryType(enum.IntEnum):
        INVALID = 0
        CUE = 1
        LOOP = 3

    def __init__(self, file_or_data: SeratoTag.FileOrData):
        self.raw_data: bytes | None
        super().__init__(file_or_data)

        self.entries: list[TrackCuesV1.Entry | TrackCuesV1.Color] = []
        if self.raw_data is not None:
            self.entries = list(self._parse(self.raw_data))

    class Entry(object):
        FORMAT = ">B4sB4s6s4sBB"
        FIELDS = (
            "start_position_set",
            "start_position",
            "end_position_set",
            "end_position",
            "field5",
            "color",
            "type",
            "is_locked",
        )
        type: bytes | str

        def __init__(self, *args):
            assert len(args) == len(self.FIELDS)
            for field, value in zip(self.FIELDS, args):
                setattr(self, field, value)

        def __repr__(self) -> str:
            return "{name}({data})".format(
                name=self.__class__.__name__,
                data=", ".join("{}={!r}".format(name, getattr(self, name)) for name in self.FIELDS),
            )

        @classmethod
        def load(cls, data: bytes):
            def _decode_bytes_32(data: bytes):
                """Decode 4 byte Serato binary format into 3 byte plain text."""
                w, x, y, z = struct.unpack("BBBB", data)
                c = (z & 0x7F) | ((y & 0x01) << 7)
                b = ((y & 0x7F) >> 1) | ((x & 0x03) << 6)
                a = ((x & 0x7F) >> 2) | ((w & 0x07) << 5)
                return struct.pack("BBB", a, b, c)

            info_size = struct.calcsize(cls.FORMAT)
            info = struct.unpack(cls.FORMAT, data[:info_size])
            entry_data = []

            start_position_set = None
            end_position_set = None
            for field, value in zip(cls.FIELDS, info):
                if field == "start_position_set":
                    assert value in (0x00, 0x7F)
                    value = value != 0x7F
                    start_position_set = value
                elif field == "end_position_set":
                    assert value in (0x00, 0x7F)
                    value = value != 0x7F
                    end_position_set = value
                elif field == "start_position":
                    assert start_position_set is not None
                    if start_position_set:
                        value = struct.unpack(">I", _decode_bytes_32(value).rjust(4, b"\x00"))[0]
                    else:
                        value = None
                elif field == "end_position":
                    assert end_position_set is not None
                    if end_position_set:
                        value = struct.unpack(">I", _decode_bytes_32(value).rjust(4, b"\x00"))[0]
                    else:
                        value = None
                elif field == "color":
                    value = _decode_bytes_32(value)
                elif field == "type":
                    value = TrackCuesV1.EntryType(value)
                entry_data.append(value)

            return cls(*entry_data)

        def dump(self):
            def _encode_bytes_32(data: bytes):
                """Encode 3 byte plain text into 4 byte Serato binary format."""
                a, b, c = struct.unpack("BBB", data)
                z = c & 0x7F
                y = ((c >> 7) | (b << 1)) & 0x7F
                x = ((b >> 6) | (a << 2)) & 0x7F
                w = a >> 5
                return bytes(bytearray([w, x, y, z]))

            entry_data = []
            for field in self.FIELDS:
                value = getattr(self, field)
                if field == "start_position_set":
                    value = 0x7F if not value else 0x00
                elif field == "end_position_set":
                    value = 0x7F if not value else 0x00
                elif field == "color":
                    value = _encode_bytes_32(value)
                elif field == "start_position":
                    if value is None:
                        value = 0x7F7F7F7F
                    else:
                        value = _encode_bytes_32(struct.pack(">I", value)[1:])
                elif field == "end_position":
                    if value is None:
                        value = 0x7F7F7F7F
                    else:
                        value = _encode_bytes_32(struct.pack(">I", value)[1:])
                elif field == "type":
                    value = int(value)
                entry_data.append(value)
            return struct.pack(self.FORMAT, *entry_data)

    class Color(Entry):
        FORMAT = ">4s"
        FIELDS = ("color",)

    def _parse(self, data: bytes):
        fp = io.BytesIO(data)
        self._check_version(fp.read(self.VERSION_LEN))

        num_entries = struct.unpack(">I", fp.read(4))[0]
        for i in range(num_entries):  # pylint: disable=unused-variable
            entry_data = fp.read(0x16)
            assert len(entry_data) == 0x16

            entry = TrackCuesV1.Entry.load(entry_data)
            yield entry

        yield TrackCuesV1.Color.load(fp.read())

    def _dump(self):
        entries = self.entries
        data = self._pack_version()
        num_entries = len(entries) - 1
        data += struct.pack(">I", num_entries)
        for entry_data in entries:
            data += entry_data.dump()
        self.raw_data = data


if __name__ == "__main__":
    import argparse
    import ast
    import configparser
    import math
    import subprocess
    import tempfile

    from serato_tools.utils.ui import get_hex_editor, get_text_editor, ui_ask

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    try:
        tags = TrackCuesV1(args.file)
    except HeaderNotFoundError:
        with open(args.file, mode="rb") as fp:
            data = fp.read()
        tags = TrackCuesV1(data)

    if args.edit:
        text_editor = get_text_editor()
        hex_editor = get_hex_editor()

    entries: list[TrackCuesV1.Entry] = tags.entries
    new_entries: list[TrackCuesV1.Entry] = []
    width = math.floor(math.log10(len(entries))) + 1
    action: str | None = None
    for entry_index, entry in enumerate(entries):
        if args.edit:
            if action not in ("q", "_"):
                print("{:{}d}: {!r}".format(entry_index, width, entry))
                action = ui_ask(
                    "Edit this entry",
                    {
                        "y": "edit this entry",
                        "n": "do not edit this entry",
                        "q": "quit; do not edit this entry or any of the remaining ones",
                        "a": "edit this entry and all later entries in the file",
                        "b": "edit raw bytes",
                        "r": "remove this entry",
                    },
                    default="n",
                )

            if action in ("y", "a", "b"):
                while True:
                    with tempfile.NamedTemporaryFile() as f:
                        if action == "b":
                            f.write(entry.dump())
                            editor = hex_editor  # pyright: ignore[reportPossiblyUnboundVariable]
                        else:
                            if action == "a":
                                entries_to_edit = (
                                    (
                                        "{:{}d}: {}".format(
                                            i,
                                            width,
                                            (repr(e.type) if e.__class__ == TrackCuesV1.Entry else "Color"),
                                        ),
                                        e,
                                    )
                                    for i, e in enumerate(entries[entry_index:], start=entry_index)
                                )
                            else:
                                entries_to_edit = (
                                    (
                                        (repr(entry.type) if entry.__class__ == TrackCuesV1.Entry else "Color"),
                                        entry,
                                    ),
                                )

                            for section, e in entries_to_edit:
                                f.write("[{}]\n".format(section).encode())
                                for field in e.FIELDS:
                                    value = getattr(e, field)
                                    if field == "type":
                                        value = int(value)
                                    f.write(
                                        "{}: {!r}\n".format(
                                            field,
                                            value,
                                        ).encode()
                                    )
                                f.write(b"\n")
                            editor = text_editor  # pyright: ignore[reportPossiblyUnboundVariable]
                        f.flush()
                        status = subprocess.call((editor, f.name))
                        f.seek(0)
                        output = f.read()

                    if status != 0:
                        if (
                            ui_ask(
                                "Command failed, retry",
                                {
                                    "y": "edit again",
                                    "n": "leave unchanged",
                                },
                            )
                            == "n"
                        ):
                            break
                    else:
                        try:
                            if action != "b":
                                cp = configparser.ConfigParser()
                                cp.read_string(output.decode())
                                sections = tuple(sorted(cp.sections()))
                                if action != "a":
                                    assert len(sections) == 1

                                results = []
                                for section in sections:
                                    l, s, r = section.partition(": ")
                                    name = r if s else l
                                    cls = TrackCuesV1.Color if name == "Color" else TrackCuesV1.Entry
                                    e = cls(
                                        *(
                                            ast.literal_eval(
                                                cp.get(section, field),
                                            )
                                            for field in cls.FIELDS
                                        )
                                    )
                                    results.append(cls.load(e.dump()))
                            else:
                                results = [entry.load(output)]
                        except Exception as e:  # pylint: disable=broad-exception-caught
                            print(str(e))
                            if (
                                ui_ask(
                                    "Content seems to be invalid, retry",
                                    {
                                        "y": "edit again",
                                        "n": "leave unchanged",
                                    },
                                )
                                == "n"
                            ):
                                break
                        else:
                            for i, e in enumerate(results, start=entry_index):
                                print("{:{}d}: {!r}".format(i, width, e))
                            subaction = ui_ask(
                                "Above content is valid, save changes",
                                {
                                    "y": "save current changes",
                                    "n": "discard changes",
                                    "e": "edit again",
                                },
                                default="y",
                            )
                            if subaction == "y":
                                new_entries.extend(results)
                                if action == "a":
                                    action = "_"
                                break
                            elif subaction == "n":
                                if action == "a":
                                    action = "q"
                                new_entries.append(entry)
                                break
            elif action in ("r", "_"):
                continue
            else:
                new_entries.append(entry)
        else:
            print("{:{}d}: {!r}".format(entry_index, width, entry))

    if args.edit:
        if new_entries == entries:
            print("No changes made.")
        else:
            tags.entries = new_entries
            tags._dump()  # pylint: disable=protected-access
            if tags.tagfile:
                tags.save()
            else:
                with open(args.file, mode="wb") as fp:
                    fp.write(tags.raw_data or b"")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import base64
import os
import sys

import mutagen._file
import mutagen.aiff
import mutagen.flac
import mutagen.mp3
import mutagen.mp4
import mutagen.oggvorbis

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.track_tags import SeratoTrack
from serato_tools.utils import logger


def get_serato_tagdata(tagfile: mutagen._file.FileType, decode: bool = False):
    if not tagfile:
        raise Exception("tagfile is None")
    if not tagfile.tags:
        raise Exception("no tags")

    if isinstance(tagfile, (mutagen.mp3.MP3, mutagen.aiff.AIFF)):
        tf = SeratoTrack(tagfile)
        for tagname in tagfile.tags.keys():
            tagname = str(tagname)
            if tagname.startswith("GEOB:Serato"):
                tagvalue = tf._get_geob(tagname.lstrip("GEOB:"))  # pylint: disable=protected-access
                if not tagvalue:
                    raise ValueError(f"no value for {tagname}")
                yield tagname[5:], tagvalue

    elif isinstance(tagfile, (mutagen.flac.FLAC, mutagen.mp4.MP4)):
        for tagname, tagvalue in tagfile.tags.items():  # type: ignore
            tagname = str(tagname)
            if not (tagname.startswith("serato_") or tagname.startswith("----:com.serato.dj:")):
                continue

            tagvalue = tagvalue[0]
            encoded_data: bytes
            if isinstance(tagfile, mutagen.flac.FLAC):
                if not isinstance(tagvalue, str):
                    raise TypeError("unexpected type")
                encoded_data = tagvalue.encode("utf-8")
            else:
                encoded_data = bytes(tagvalue)

            fixed_data: bytes = encoded_data

            length = len(fixed_data.splitlines()[-1])
            if length % 4 == 1:
                fixed_data += b"A=="
            elif length % 4 == 2:
                fixed_data += b"=="
            elif length % 4 == 3:
                fixed_data += b"="
            data = base64.b64decode(fixed_data)

            if not data.startswith(b"application/octet-stream\0"):
                logger.error(f"Failed to parse tag: {tagname}")
                continue
            fieldname_endpos = data.index(b"\0", 26)
            fieldname = data[26:fieldname_endpos].decode()
            fielddata = data[fieldname_endpos + 1 :]
            yield fieldname, fielddata if decode else encoded_data

    elif isinstance(tagfile, mutagen.oggvorbis.OggVorbis):
        for tagname, tagvalue in tagfile.tags.items():  # type: ignore
            tagname = str(tagname)
            tagvalue = tagvalue[0]
            if not tagname.startswith("serato_"):
                continue
            if not isinstance(tagvalue, str):
                raise TypeError("unexpected type")
            yield tagname, tagvalue.encode("utf-8")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("input_file")
    parser.add_argument("output_dir", nargs="?")
    parser.add_argument("-d", "--decode", action="store_true")
    args = parser.parse_args()

    # pylint: disable-next=protected-access
    tagfile: mutagen._file.FileType | None = mutagen._file.File(args.input_file)

    if not tagfile:
        raise Exception("couldn't parse file")

    for field, value in get_serato_tagdata(tagfile, decode=args.decode):
        log_str = f'field {field} { "decoded " if args.decode else "" }value: {value}'
        if args.output_dir:
            filepath = os.path.join(args.output_dir, f"{field}.octet-stream")
            print(log_str + f" (written to file {filepath})")
            with open(filepath, mode="wb") as fp:
                fp.write(value)
        else:
            print(log_str)



import os
import shutil
from typing import Optional


def ui_ask(question: str, choices: dict, default: Optional[str] = None):
    choices_str = "/".join(x.upper() if x == default else x for x in (*choices.keys(), "?"))
    text = f"{question} [{choices_str}]? "

    while True:
        answer = input(text).lower()
        if default and answer == "":
            answer = default

        if answer in choices.keys():
            return answer
        else:
            print("\n".join(f"{choice} - {desc}" for choice, desc in (*choices.items(), ("?", "print help"))))


def get_text_editor():
    text_editor = shutil.which(os.getenv("EDITOR", "vi"))
    if not text_editor:
        raise Exception("No suitable EDITOR found.")
    return text_editor


def get_hex_editor():
    hex_editor = shutil.which(os.getenv("HEXEDITOR", "bvi"))
    if not hex_editor:
        raise Exception("No suitable HEXEDITOR found.")
    return hex_editor



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.track_cues_v2 import TrackCuesV2


class TestCase(unittest.TestCase):
    def test_parse_and_dump(self):
        with open(os.path.abspath("test/data/track_cues_v2.bin"), mode="rb") as fp:
            file_data = fp.read()
        tags = TrackCuesV2(file_data)
        self.assertEqual(
            [str(e) for e in tags.entries],
            [
                "ColorEntry(field1=b'\\x00', color=b'\\x99\\xff\\x99')",
                "CueEntry(field1=b'\\x00', index=0, position=29, field4=b'\\x00', color=b'\\x88\\x00\\xcc', field6=b'\\x00\\x00', name='')",
                "CueEntry(field1=b'\\x00', index=1, position=12829, field4=b'\\x00', color=b'\\x00\\xcc\\x00', field6=b'\\x00\\x00', name='LYRICS')",
                "CueEntry(field1=b'\\x00', index=2, position=51229, field4=b'\\x00', color=b'\\xcc\\xcc\\x00', field6=b'\\x00\\x00', name='')",
                "CueEntry(field1=b'\\x00', index=3, position=64029, field4=b'\\x00', color=b'\\xcc\\x88\\x00', field6=b'\\x00\\x00', name='')",
                "CueEntry(field1=b'\\x00', index=4, position=89629, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='')",
                "CueEntry(field1=b'\\x00', index=5, position=102429, field4=b'\\x00', color=b'\\xcc\\xcc\\x00', field6=b'\\x00\\x00', name='LYRICS')",
                "CueEntry(field1=b'\\x00', index=6, position=153629, field4=b'\\x00', color=b'\\xcc\\x88\\x00', field6=b'\\x00\\x00', name='')",
                "CueEntry(field1=b'\\x00', index=7, position=204829, field4=b'\\x00', color=b'\\x88\\x00\\xcc', field6=b'\\x00\\x00', name='')",
                "BpmLockEntry(enabled=False)",
            ],
            "parsed entries",
        )
        tags._dump()
        self.assertEqual(tags.raw_data, file_data, "dump")

        tags.modify_entries(
            {
                "cues": [
                    {"field": "color", "func": lambda val: TrackCuesV2.CueColors.RED},
                    {"field": "name", "func": lambda val: "NEW" if val == "" else None},
                ],
                "color": [
                    {"field": "color", "func": lambda val: TrackCuesV2.TrackColors.ORANGE},
                ],
            },
        )

        self.assertEqual(
            [str(e) for e in tags.entries],
            [
                "ColorEntry(field1=b'\\x00', color=b'\\xff\\xbb\\x99')",
                "CueEntry(field1=b'\\x00', index=0, position=29, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='NEW')",
                "CueEntry(field1=b'\\x00', index=1, position=12829, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='LYRICS')",
                "CueEntry(field1=b'\\x00', index=2, position=51229, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='NEW')",
                "CueEntry(field1=b'\\x00', index=3, position=64029, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='NEW')",
                "CueEntry(field1=b'\\x00', index=4, position=89629, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='NEW')",
                "CueEntry(field1=b'\\x00', index=5, position=102429, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='LYRICS')",
                "CueEntry(field1=b'\\x00', index=6, position=153629, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='NEW')",
                "CueEntry(field1=b'\\x00', index=7, position=204829, field4=b'\\x00', color=b'\\xcc\\x00\\x00', field6=b'\\x00\\x00', name='NEW')",
                "BpmLockEntry(enabled=False)",
            ],
            "modified entries",
        )
        tags._dump()

        with open(os.path.abspath("test/data/track_cues_v2_modified.bin"), mode="rb") as fp:
            expected_modified_data = fp.read()
        self.assertEqual(tags.raw_data, expected_modified_data, "modified data dump")



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.track_waveform import TrackWaveform


class TestCase(unittest.TestCase):
    def test_parse(self):
        with open(os.path.abspath("test/data/track_waveform.bin"), mode="rb") as fp:
            file_data = fp.read()
        tags = TrackWaveform(file_data)
        self.assertEqual(tags.raw_data, file_data, "raw_data read")
        with open(os.path.abspath("test/data/track_waveform_parsed.bin"), mode="rb") as fp:
            expected_parsed_data = fp.read()
        self.assertEqual(
            b"".join(bytes(x) for x in list(tags.data)),
            expected_parsed_data,
            "parsed data",
        )



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.track_autotags import TrackAutotags


class TestCase(unittest.TestCase):
    def test_parse_and_dump(self):
        with open(os.path.abspath("test/data/track_autotags.bin"), mode="rb") as fp:
            file_data = fp.read()
        tags = TrackAutotags(file_data)
        self.assertEqual(tags.raw_data, file_data, "raw_data read")
        self.assertEqual(tags.bpm, 75.0, "parsed bpm")
        self.assertEqual(tags.autogain, -5.074, "parsed autogain")
        self.assertEqual(tags.gaindb, 0.0, "parsed gaindb")
        tags._dump()
        self.assertEqual(tags.raw_data, file_data, "dump")



import sys
import os
import filecmp
import shutil
import glob
import re
import platform
from typing import Optional, cast

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.database_v2 import DatabaseV2
from serato_tools.crate import Crate
from serato_tools.smart_crate import SmartCrate
from serato_tools.utils import (
    logger,
    SERATO_DIR_NAME,
    SERATO_DIR as LOCAL_SERATO_DIR,
    SERATO_DRIVE as LOCAL_SERATO_DRIVE,
)

# TODO: put in folders if there are different locations locally for the same basename


def _uniq_by_basename(paths: list[str]):
    basenames: list[str] = []
    return_paths: list[str] = []
    for path in paths:
        basename = os.path.basename(path)
        if basename not in basenames:
            basenames.append(basename)
            return_paths.append(path)
    return return_paths


def copy_crates_to_usb(
    crate_files: list[str], dest_drive_dir: str, dest_tracks_dir: str, root_crate: Optional[str] = None
):
    if not os.path.isdir(dest_drive_dir):
        logger.error(f"destination drive directory does not exist: {dest_drive_dir}")

    DEST_SERATO_DIR = os.path.join(dest_drive_dir, SERATO_DIR_NAME)

    # TODO: merge with existing, instead of replacing
    if os.path.isdir(DEST_SERATO_DIR):
        shutil.rmtree(DEST_SERATO_DIR)

    # copy the crate files, and get the filenames from them
    tracks_to_copy: list[str] = []

    crate_dir = os.path.join(DEST_SERATO_DIR, Crate.DIR)
    os.makedirs(crate_dir, exist_ok=True)

    def change_track_path(track_path: str):
        return Crate.get_relative_path(os.path.join(dest_tracks_dir, os.path.basename(track_path)))

    def modify_crate_track(track: Crate.Track) -> Crate.Track:
        tracks_to_copy.append(track.relpath)
        track.set_path(change_track_path(track.relpath))
        return track

    for crate_file in crate_files:
        if os.path.isdir(crate_file):
            continue
        elif crate_file.endswith(Crate.EXTENSION):
            crate = Crate(crate_file)
        elif crate_file.endswith(SmartCrate.EXTENSION):
            crate = SmartCrate(crate_file)
        else:
            raise ValueError(
                f"not a crate file, should end with {Crate.EXTENSION} or {SmartCrate.EXTENSION}: {crate_file}"
            )

        crate.modify_tracks(modify_crate_track)
        crate.remove_duplicates()

        crate_filename = os.path.basename(crate_file)
        if root_crate is not None:
            crate_filename = root_crate + "%%" + crate_filename
        new_crate_file = os.path.join(crate_dir, crate_filename)
        crate.save(new_crate_file)

        # need to just save smartCrates as a .crate instead of a .scrate, can't do smart crate on USB.
        if isinstance(crate, SmartCrate):
            existing = new_crate_file
            new = existing.replace("≫≫", "%%").replace(SmartCrate.EXTENSION, Crate.EXTENSION)
            os.replace(existing, new)
            new_crate_file = new
        logger.info(f"copied crate {new_crate_file}")

    # create the db file
    db = DatabaseV2()
    tracks_to_copy = [os.path.normpath(f) for f in tracks_to_copy]
    tracks_to_copy = _uniq_by_basename(tracks_to_copy)
    tracks_to_copy_basenames = [os.path.basename(f) for f in tracks_to_copy]

    def modify_db_track(track: DatabaseV2.Track) -> DatabaseV2.Track:
        track.set_path(os.path.join(dest_tracks_dir, os.path.basename(track.relpath)))
        track.set_value(DatabaseV2.Fields.PLAYED, False)
        return track

    db.filter_tracks(lambda track: os.path.basename(track.relpath) in tracks_to_copy_basenames)
    db.modify_tracks(modify_db_track)
    db.remove_duplicates()

    new_db_file = os.path.join(DEST_SERATO_DIR, DatabaseV2.FILENAME)
    db.save(new_db_file)
    logger.info(f"created database file {new_db_file}")

    # copy crate order file, modify if needed
    NEW_ORDER_FILE = "neworder.pref"
    with open(os.path.join(LOCAL_SERATO_DIR, NEW_ORDER_FILE), "r", encoding="utf-8") as f:
        lines = f.read().splitlines()

    if root_crate:
        CRATE_KEY = "[crate]"
        new_lines = []
        for line in lines:
            if line.startswith(CRATE_KEY):
                value = line[len(CRATE_KEY) :]
                if value != "Stems":
                    line = f"{CRATE_KEY}{root_crate}%%{value}"
            new_lines.append(line)
        lines = new_lines

    DEST_NEW_ORDER_FILEPATH = os.path.join(DEST_SERATO_DIR, NEW_ORDER_FILE)
    with open(DEST_NEW_ORDER_FILEPATH, "w", encoding="utf-8") as f:
        f.writelines(lines)
        logger.info(f"copied order file {DEST_NEW_ORDER_FILEPATH}")

    # copy stems crate
    LOCAL_STEMS_CRATE = os.path.join(LOCAL_SERATO_DIR, Crate.SERATO_STEMS_CRATE_PATH)
    if os.path.exists(LOCAL_STEMS_CRATE):
        DEST_STEMS_CRATE = os.path.join(DEST_SERATO_DIR, Crate.SERATO_STEMS_CRATE_PATH)
        os.makedirs(os.path.dirname(DEST_STEMS_CRATE), exist_ok=True)
        stems_crate = Crate(LOCAL_STEMS_CRATE)

        def modify_stems_crate_track(track: Crate.Track) -> Crate.Track:
            track.set_path(change_track_path(track.relpath))
            return track

        stems_crate.filter_tracks(lambda track: os.path.basename(track.relpath) in tracks_to_copy_basenames)
        stems_crate.modify_tracks(modify_stems_crate_track)
        stems_crate.remove_duplicates()
        stems_crate.save(DEST_STEMS_CRATE)
        logger.info(f"copied crate {DEST_STEMS_CRATE}")

    # copy files
    logger.info("copying files over...")

    root = LOCAL_SERATO_DRIVE + os.sep if LOCAL_SERATO_DRIVE else os.sep
    tracks_to_copy = [os.path.join(root, t) for t in tracks_to_copy]

    not_found: list[str] = []

    def maybe_copy(src_path: str, dst_path: str):
        if not os.path.exists(src_path):
            not_found.append(src_path)
            return
        copy = (not filecmp.cmp(src_path, dst_path, shallow=True)) if os.path.exists(dst_path) else True
        if copy:
            os.makedirs(os.path.dirname(dst_path), exist_ok=True)
            logger.info(f"copying {src_path}")
            shutil.copy2(src_path, dst_path)

    for index, src_path in enumerate(tracks_to_copy):
        percent_complete = index / float(len(tracks_to_copy)) * 100
        print(f"{percent_complete:.1f}%", end="\r", flush=True)
        maybe_copy(
            src_path=src_path,
            dst_path=os.path.join(dest_drive_dir, dest_tracks_dir, os.path.basename(src_path)),
        )

        maybe_stem_files = glob.glob(os.path.splitext(src_path)[0] + "*.serato-stems")
        if len(maybe_stem_files) > 1:
            ERROR_STR = "should not have more than 1 stem file for: " + src_path
            print(ERROR_STR)
            print("Files found:")
            print("\n".join(maybe_stem_files))
            raise Exception(ERROR_STR)
        if len(maybe_stem_files) == 1:
            maybe_copy(
                src_path=maybe_stem_files[0],
                dst_path=os.path.join(dest_drive_dir, dest_tracks_dir, os.path.basename(maybe_stem_files[0])),
            )
    print("\n")

    if len(not_found) == len(tracks_to_copy):
        logger.error("No source files found.")
        uniq_dirs = list(set(os.path.dirname(d) for d in tracks_to_copy))
        logger.error(f"Directories: \n{"\n    ".join(uniq_dirs)}")
        return

    for n in not_found:
        logger.warning(f"ERROR: does not exist - {n}")


def get_crate_files(pattern: str):
    try:
        regex = re.compile(pattern)

        def get_files_by_regex(folder_path: str) -> list[str]:
            matched_files = []
            for filename in os.listdir(folder_path):
                full_path = os.path.join(folder_path, filename)
                if os.path.isfile(full_path) and regex.search(filename):
                    matched_files.append(full_path)
            return matched_files

        func = get_files_by_regex
    except re.error:

        def get_files_by_glob(folder_path: str) -> list[str]:
            return glob.glob(pattern, root_dir=folder_path, recursive=False)

        func = get_files_by_glob

    files: list[str] = []
    for dir in [Crate.DIR, SmartCrate.DIR]:
        root_dir = os.path.join(LOCAL_SERATO_DIR, dir)
        files += [os.path.join(root_dir, f) for f in func(root_dir)]
    files = [i for i in files if os.path.isfile(i)]  # remove dirs
    return files


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--drive",
        "--drive_dir",
        dest="dest_dir",
        type=str,
        required=True,
        help='Directory of the destination drive. Example: "E" on Windows',
    )
    parser.add_argument(
        "-c",
        "--crates",
        "--crate_matcher",
        dest="crate_matcher",
        nargs="+",
        type=str,
        required=True,
        help='Glob or Regex matcher for crate and smartcrate filenames. Example: "*house*". Can pass multiple to OR them. To copy all, pass "*"',
    )
    parser.add_argument(
        "-r",
        "--root_crate",
        type=str,
        required=False,
        default=None,
        help="Not required, but is very nice when plugging your drive into another DJ's laptop. Sets all crates to be within this crate on the destination drive",
    )
    args = parser.parse_args()

    dest_dir = cast(str, args.dest_dir)

    crate_files: list[str] = []
    for cm in cast(list[str], args.crate_matcher):
        crate_files += get_crate_files(cm)

    if platform.system() == "Windows" and len(dest_dir) == 1:
        dest_dir += ":\\\\"

    copy_crates_to_usb(
        crate_files=crate_files,
        dest_drive_dir=dest_dir,
        dest_tracks_dir="Tracks",
        root_crate=args.root_crate,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# This is from this repo: https://github.com/sharst/seratopy
import os
import sys
from typing import Optional, Iterable

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.bin_file_base import SeratoBinFile


class CrateBase(SeratoBinFile):
    EXTENSION: str
    DIR: str
    DIR_PATH: str
    TRACK_PATH_KEY = SeratoBinFile.Fields.TRACK_PATH

    def _stringify_entry(self, entry: SeratoBinFile.EntryFull, indent: int = 0) -> str:
        field, fieldname, value = entry  # pylint: disable=unused-variable
        if isinstance(value, list):
            return self._stringify_entries(value, indent)
        return str(value)

    def _stringify_entries(self, entries: Iterable[SeratoBinFile.EntryFull], indent: int = 0) -> str:
        lines: list[str] = []

        for field, fieldname, value in entries:
            if isinstance(value, list):
                field_lines = [
                    f"[ {entry[0]} ({entry[1]}): {self._stringify_entry(entry, indent + 1)} ]" for entry in value
                ]
                print_val = ", ".join(field_lines)
            else:
                print_val = str(value)
            lines.append(f"{field} ({fieldname}): {print_val}")

        return "\n".join(lines)

    def print_track_paths(self, filenames_only: bool = False):
        track_paths = self.get_track_paths(include_drive=True)
        if filenames_only:
            track_paths = [os.path.splitext(os.path.basename(t))[0] for t in track_paths]
        print("\n".join(track_paths))

    def save(self, file: Optional[str] = None):
        if file is None:
            file = self.filepath

        if not file.endswith(self.EXTENSION):
            raise ValueError(f"file should end with {self.EXTENSION}: " + file)

        self._dump()
        super().save(file)

    def add_track(self, filepath: str):
        # filepath name must include the containing dir
        filepath = self.get_relative_path(filepath)

        if filepath in self.get_track_paths():
            return

        self.entries.append((CrateBase.Fields.TRACK, [(CrateBase.Fields.TRACK_PATH, filepath)]))

    def add_tracks_from_dir(self, dir: str, replace: bool = False):
        dir_tracks = [self.get_relative_path(os.path.join(dir, t)) for t in os.listdir(dir)]

        if replace:
            for track in self.get_track_paths():
                if track not in dir_tracks:
                    self.remove_track(track)

        for track in dir_tracks:
            self.add_track(track)

    @classmethod
    def get_crate_files(cls, file_or_dir: str | None = None) -> list[str]:
        if file_or_dir is None:
            file_or_dir = cls.DIR_PATH

        crate_paths: list[str] = []
        if os.path.isdir(file_or_dir):

            for fname in os.listdir(file_or_dir):
                path = os.path.normpath(os.path.join(file_or_dir, fname))
                if os.path.isfile(path) and path.endswith(cls.EXTENSION):
                    crate_paths.append(path)
        elif os.path.isfile(file_or_dir):
            crate_paths = [file_or_dir]
        else:
            raise FileNotFoundError("does not exist: " + file_or_dir)

        return crate_paths



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import os
import sys
from typing import Optional

from mutagen.mp3 import HeaderNotFoundError

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.track_tags import SeratoTag
from serato_tools.utils import logger


class TrackAutotags(SeratoTag):
    GEOB_KEY = "Serato Autotags"
    VERSION = (0x01, 0x01)

    def __init__(self, file_or_data: SeratoTag.FileOrData):
        self.raw_data: bytes | None = None
        super().__init__(file_or_data)

        self.bpm: float | None = None
        self.autogain: float | None = None
        self.gaindb: float | None = None

        if self.raw_data is not None:
            self.bpm, self.autogain, self.gaindb = self._parse(self.raw_data)

    def __str__(self) -> str:
        return f"bpm: {self.bpm}\nautogain: {self.autogain}\ngaindb: {self.gaindb}"

    def _parse(self, data: bytes):
        fp = io.BytesIO(data)
        self._check_version(fp.read(self.VERSION_LEN))

        def get_value():
            value = self._readbytes(fp).decode("ascii")
            return float(value)

        bpm = get_value()
        autogain = get_value()
        gaindb = get_value()
        return bpm, autogain, gaindb

    def _dump(self):
        data: bytes = self._pack_version()
        for value, decimals in ((self.bpm, 2), (self.autogain, 3), (self.gaindb, 3)):
            data += "{:.{}f}".format(value, decimals).encode("ascii")
            data += b"\x00"
        self.raw_data = data

    def set(
        self,
        bpm: Optional[float] = None,
        autogain: Optional[float] = None,
        gaindb: Optional[float] = None,
    ):
        if bpm is not None:
            self.bpm = bpm
        if autogain is not None:
            self.autogain = autogain
        if gaindb is not None:
            self.gaindb = gaindb
        self._dump()


if __name__ == "__main__":
    import argparse
    import configparser
    import subprocess
    import tempfile

    from serato_tools.utils.ui import get_text_editor

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    try:
        tags = TrackAutotags(args.file)
    except HeaderNotFoundError:
        with open(args.file, mode="rb") as fp:
            data = fp.read()
        tags = TrackAutotags(data)

    if args.edit:
        editor = get_text_editor()

        with tempfile.NamedTemporaryFile() as f:
            f.write(str(tags).encode("ascii"))
            f.flush()
            status = subprocess.call((editor, f.name))
            f.seek(0)
            output = f.read()

        if status != 0:
            ERROR_STR = f"Command executation failed with status: {status}"
            print(ERROR_STR, file=sys.stderr)
            logger.error(ERROR_STR)
            raise Exception(ERROR_STR)

        cp = configparser.ConfigParser()
        try:
            SECTION = "Autotags"
            cp.read_string(f"[{SECTION}]\n" + output.decode())
            bpm = cp.getfloat(SECTION, "bpm")
            autogain = cp.getfloat(SECTION, "autogain")
            gaindb = cp.getfloat(SECTION, "gaindb")
        except Exception:
            ERROR_STR = "Invalid input, no changes made"
            print(ERROR_STR, file=sys.stderr)
            logger.error(ERROR_STR)
            raise

        tags.set(bpm=bpm, autogain=autogain, gaindb=gaindb)
        if tags.tagfile:
            tags.save()
        else:
            with open(args.file, mode="wb") as fp:
                fp.write(tags.raw_data or b"")
    else:
        print(str(tags))



# pylint: disable=protected-access
import unittest
import os
import json

from src.serato_tools.database_v2 import DatabaseV2


class TestCase(unittest.TestCase):
    def test_format_filepath(self):
        self.assertEqual(
            DatabaseV2.get_relative_path("C:\\Music\\DJ Tracks\\Zeds Dead - In The Beginning.mp3"),
            "Music/DJ Tracks/Zeds Dead - In The Beginning.mp3",
        )
        self.assertEqual(
            DatabaseV2.get_relative_path("Music\\DJ Tracks\\Zeds Dead - In The Beginning.mp3"),
            "Music/DJ Tracks/Zeds Dead - In The Beginning.mp3",
        )
        self.assertEqual(
            DatabaseV2.get_relative_path("C:/Music/DJ Tracks/Tripp St. - Enlighten.mp3"),
            "Music/DJ Tracks/Tripp St. - Enlighten.mp3",
        )
        self.assertEqual(
            DatabaseV2.get_relative_path("Music/DJ Tracks/Tripp St. - Enlighten.mp3"),
            "Music/DJ Tracks/Tripp St. - Enlighten.mp3",
        )

    def test_parse_and_dump(self):
        file = os.path.abspath("test/data/database_v2_test.bin")
        with open(file, mode="rb") as fp:
            file_data = fp.read()

        db = DatabaseV2(file)

        self.maxDiff = None
        self.assertEqual(db.raw_data, file_data, "raw_data read")

        with open("test/data/database_v2_test_output.txt", "r", encoding="utf-16") as f:
            expected = f.read()
            self.assertEqual(db.__str__(), expected, "parse")

        db._dump()
        self.assertEqual(db.raw_data, file_data, "raw_data read")

    def test_parse_and_modify(self):
        db = DatabaseV2(os.path.abspath("test/data/database_v2_test.bin"))

        self.maxDiff = None

        with open("test/data/database_v2_test_output.txt", "r", encoding="utf-16") as f:
            expected = f.read()
            self.assertEqual(db.__str__(), expected, "parse")

        original_entries = db.entries
        original_raw_data = db.raw_data
        db.modify([])
        self.assertEqual(db.entries, original_entries, "was not modified")
        self.assertEqual(db.raw_data, original_raw_data, "was not modified")
        self.assertEqual(db.__str__(), expected, "was not modified")

        new_time = int(1735748100)
        db.modify(
            [
                {"field": DatabaseV2.Fields.DATE_ADDED_U, "func": lambda *args: new_time},
                {"field": DatabaseV2.Fields.DATE_ADDED_T, "func": lambda *args: str(new_time)},
                {"field": DatabaseV2.Fields.GROUPING, "func": lambda *args: "NEW_GROUPING"},
            ]
        )
        with open("test/data/database_v2_test_modified_output.txt", "r", encoding="utf-16") as f:
            self.assertEqual(db.__str__(), f.read(), "was modified correctly")
        with open("test/data/database_v2_test_modified_output.bin", "rb") as f:
            self.assertEqual(db.raw_data, f.read(), "was modified correctly")

        db.modify(
            [
                {
                    "field": DatabaseV2.Fields.GENRE,
                    "func": lambda *args: "NEW_GENRE",
                    "files": [
                        "Users\\bvand\\Music\\DJ Tracks\\Zeds Dead - In The Beginning.mp3",
                        "C:/Users/bvand/Music/DJ Tracks/Tripp St. - Enlighten.mp3",
                    ],
                },
            ]
        )
        with open("test/data/database_v2_test_modified_output_2.txt", "r", encoding="utf-8") as f:
            self.assertEqual(db.__str__(), f.read(), "was modified correctly, given files")
        with open("test/data/database_v2_test_modified_output_2.bin", "rb") as f:
            self.assertEqual(db.raw_data, f.read(), "was modified correctly, given files")

    def test_dedupe(self):
        db = DatabaseV2(os.path.abspath("test/data/database_v2_duplicates.bin"))

        with open("test/data/database_v2_duplicates_output.txt", "r", encoding="utf-8") as f:
            self.assertEqual(db.__str__(), f.read(), "original")

        db.remove_duplicates()

        with open("test/data/database_v2_duplicates_output_deduped.txt", "r", encoding="utf-8") as f:
            self.assertEqual(db.__str__(), f.read(), "deduped")

    def test_to_json_object(self):
        db = DatabaseV2(os.path.abspath("test/data/database_v2_test.bin"))

        with open("test/data/database_v2_json.json", "r", encoding="utf-8") as f:
            self.assertEqual(db.to_json_object(), json.loads(f.read()))

    def test_from_json_object(self):
        db_from_json = DatabaseV2(os.path.abspath("test/data/database_v2_json.json"))
        db_from_bin = DatabaseV2(os.path.abspath("test/data/database_v2_test.bin"))

        self.assertEqual(db_from_json.__str__(), db_from_bin.__str__())

        json_object = db_from_bin.to_json_object()
        db_from_json.from_json_object(json_object)

        self.assertEqual(db_from_json.__str__(), db_from_bin.__str__())



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.track_beatgrid import TrackBeatgrid


class TestCase(unittest.TestCase):
    def test_parse_and_dump(self):
        with open(os.path.abspath("test/data/track_beatgrid.bin"), mode="rb") as fp:
            file_data = fp.read()
        tags = TrackBeatgrid(file_data)
        self.assertEqual(tags.raw_data, file_data, "raw_data read")
        assert tags.entries is not None
        self.assertEqual(
            tags.entries,
            [
                TrackBeatgrid.TerminalBeatgridMarker(position=0.029895611107349396, bpm=75.0),
                TrackBeatgrid.Footer(unknown=0),
            ],
            "parsed entries",
        )
        tags._dump()
        self.assertEqual(tags.raw_data, file_data, "dump")



#!/usr/bin/python
# This is from this repo: https://github.com/sharst/seratopy
import os
import sys

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.crate_base import CrateBase
from serato_tools.utils import SERATO_DIR


class Crate(CrateBase):
    TESTED_VERSIONS = ["1.0/Serato ScratchLive Crate"]
    EXTENSION = ".crate"
    DIR = "Subcrates"
    DIR_PATH = os.path.join(SERATO_DIR, DIR)
    SERATO_STEMS_CRATE_PATH = os.path.join(DIR, "Serato Stems", "Stems.crate")

    DEFAULT_ENTRIES = [
        (CrateBase.Fields.VERSION, "1.0/Serato ScratchLive Crate"),
        (CrateBase.Fields.SORTING, [(CrateBase.Fields.COLUMN_NAME, "key"), (CrateBase.Fields.REVERSE_ORDER, False)]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "song"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "playCount"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "artist"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "bpm"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "key"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "album"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "length"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "comment"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
        (CrateBase.Fields.COLUMN, [(CrateBase.Fields.COLUMN_NAME, "added"), (CrateBase.Fields.COLUMN_WIDTH, "0")]),
    ]


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file_or_dir", nargs="?", help="Set the crate file, or a directory to do all files in it")
    parser.add_argument("--all", action="store_true", help="Perform actions on all crates.")
    parser.add_argument("-l", "--list_tracks", action="store_true", help="Only list tracks")
    parser.add_argument("-f", "--filenames_only", action="store_true", help="Only list track filenames")
    parser.add_argument("--find_missing", action="store_true", help="List files that do not exist")
    args = parser.parse_args()

    if args.all:
        args.file_or_dir = Crate.DIR_PATH

    if not args.file_or_dir:
        print(f"must pass a file or dir, or --all!\nfiles in {Crate.DIR_PATH}:")
        print("\n".join(Crate.get_crate_files()))
        sys.exit()

    for crate_path in Crate.get_crate_files(args.file_or_dir):
        crate = Crate(crate_path)
        if args.find_missing:
            crate.find_missing()
        elif args.list_tracks or args.filenames_only:
            crate.print_track_paths(filenames_only=args.filenames_only)
        else:
            print(crate)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import sys

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.bin_file_base import SeratoBinFile
from serato_tools.utils import logger, SERATO_DIR


class DatabaseV2(SeratoBinFile):
    TESTED_VERSIONS = ["2.0/Serato Scratch LIVE Database"]
    FILENAME = "database V2"
    DEFAULT_DATABASE_FILE = os.path.join(SERATO_DIR, FILENAME)
    TRACK_PATH_KEY = SeratoBinFile.Fields.FILE_PATH

    DEFAULT_ENTRIES = [
        (SeratoBinFile.Fields.VERSION, "2.0/Serato Scratch LIVE Database"),
    ]

    def __init__(self, file: str = DEFAULT_DATABASE_FILE):
        if not os.path.exists(file):
            raise FileNotFoundError(f"file does not exist: {file}")
        super().__init__(file=file)

    def rename_track_file(self, src: str, dest: str):
        """
        This renames the file path, and also changes the path in the database to point to the new filename, so that
        the renamed file is not missing in the library.
        """
        try:
            os.rename(src=src, dst=dest)
            logger.info(f"renamed {src} to {dest}")
        except FileExistsError:
            # can't just do os.path.exists, doesn't pick up case changes for certain filesystems
            logger.error(f"File already exists with change: {src}")
            return
        self.change_track_path(src, dest)
        self.save()

        from serato_tools.crate import Crate

        for crate_path in Crate.get_crate_files():
            crate = Crate(crate_path)
            crate.change_track_path(src, dest)
            crate.save()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs="?", default=DatabaseV2.DEFAULT_DATABASE_FILE)
    parser.add_argument("--find_missing", action="store_true", help="List files that do not exist")
    args = parser.parse_args()

    db = DatabaseV2(args.file)

    if args.find_missing:
        # TODO: actually look for that missing flag.
        db.find_missing()
    else:
        print(db)



# pylint: disable=protected-access
import unittest
import os

from mutagen.mp3 import MP3


from src.serato_tools.track_cues_v2 import TrackCuesV2
from src.serato_tools.track_cues_v1 import TrackCuesV1
from src.serato_tools.track_waveform import TrackWaveform
from src.serato_tools.track_autotags import TrackAutotags
from src.serato_tools.utils.track_tags import SeratoTrack


class TestCase(unittest.TestCase):
    def test_file_parse(self):
        tagfile = MP3(os.path.abspath("test/data/test_mp3.mp3"))
        TrackCuesV2(tagfile).delete()
        TrackCuesV1(tagfile).delete()
        TrackWaveform(tagfile).delete()
        TrackAutotags(tagfile).delete()

        self.assertIsNone(TrackCuesV2(tagfile)._get_geob())
        self.assertIsNone(TrackCuesV1(tagfile)._get_geob())
        self.assertIsNone(SeratoTrack(tagfile)._get_geob(TrackWaveform.GEOB_KEY))
        self.assertIsNone(TrackAutotags(tagfile)._get_geob())



import os
import io
import re
import struct
import base64
from enum import StrEnum
import json
from typing import Iterable, TypedDict, Generator, Optional, cast, Callable, Pattern, NotRequired

from serato_tools.utils import get_enum_key_from_value, logger, SERATO_DRIVE, DataTypeError, DeeplyNestedListError


class SeratoBinFile:

    class Fields(StrEnum):
        # Database & Crate
        VERSION = "vrsn"
        TRACK = "otrk"
        # Database - Track
        FILE_TYPE = "ttyp"
        FILE_PATH = "pfil"
        TITLE = "tsng"
        ARTIST = "tart"
        ALBUM = "talb"
        GENRE = "tgen"
        LENGTH = "tlen"
        BITRATE = "tbit"
        SAMPLE_RATE = "tsmp"
        SIZE = "tsiz"
        BPM = "tbpm"
        KEY = "tkey"
        TIME = "utme"
        GROUPING = "tgrp"
        PUBLISHER = "tlbl"
        COMPOSER = "tcmp"
        YEAR = "ttyr"
        DATE_ADDED_T = "tadd"
        DATE_ADDED_U = "uadd"
        BEATGRID_LOCKED = "bbgl"
        CORRUPT = "bcrt"
        MISSING = "bmis"
        HAS_STEMS = "bstm"
        PLAYED = "bply"
        # Crates
        SORTING = "osrt"
        REVERSE_ORDER = "brev"
        COLUMN = "ovct"
        COLUMN_NAME = "tvcn"
        COLUMN_WIDTH = "tvcw"
        TRACK_PATH = "ptrk"
        # Smart Crates
        SMARTCRATE_RULE = "rurt"
        SMARTCRATE_LIVE_UPDATE = "rlut"
        SMARTCRATE_MATCH_ALL = "rart"
        RULE_VALUE_TEXT = "trpt"
        RULE_VALUE_DATE = "trtt"
        RULE_VALUE_INTEGER = "urpt"
        RULE_COMPARISON = "trft"
        RULE_FIELD = "urkt"

    FIELDS = list(f.value for f in Fields)

    type ParsedField = Fields | str
    type BasicValue = str | bytes | int | bool

    type Entry = tuple[ParsedField, "SeratoBinFile.Value"]
    type EntryList = list[Entry]
    type Value = BasicValue | EntryList

    type EntryFull = tuple[ParsedField, str, "SeratoBinFile.EntryFullValue"]
    type EntryFullList = list[EntryFull]
    type EntryFullValue = BasicValue | EntryFullList

    type ValueOrNone = Value | None

    TESTED_VERSIONS: list[str]
    TRACK_PATH_KEY: Fields
    DEFAULT_ENTRIES: EntryList

    def __init__(self, file: str):
        self.filepath = os.path.abspath(file)

        self.raw_data: bytes
        self.entries: SeratoBinFile.EntryList
        self.version: str

        for key in ["TESTED_VERSIONS", "TRACK_PATH_KEY", "DEFAULT_ENTRIES"]:
            if not hasattr(self, key):
                raise AttributeError(f"need to set {key} in subclass")

        if os.path.exists(file):
            if file.lower().endswith(".json"):
                with open(file, "r", encoding="utf-8") as f:
                    self.from_json_object(json.load(f))
            else:
                with open(file, "rb") as f:
                    self.raw_data = f.read()
                    self.entries = list(self._parse_item(self.raw_data))
        else:
            logger.warning(f"File does not exist: {file}. Using default data to create an empty item.")
            self.entries = self.DEFAULT_ENTRIES
            version_entry = self.DEFAULT_ENTRIES[0]
            default_version = version_entry[1]
            if not isinstance(default_version, str):
                raise DataTypeError(default_version, str, version_entry[0])
            if default_version not in self.TESTED_VERSIONS:
                raise ValueError("version in DEFAULT_ENTRIES is not in TESTED_VERSIONS")
            self.version = default_version
            self._dump()

        if not hasattr(self, "version"):
            raise AttributeError("version not set after parsing file")

    def __str__(self) -> str:
        return self._stringify_entries(self.get_entries())

    def _stringify_entries(self, entries: Iterable[EntryFull], indent: int = 0) -> str:
        lines: list[str] = []
        indent_str = "    " * indent

        for field, fieldname, value in entries:
            if isinstance(value, list):
                lines.append(f"{indent_str}{field} ({fieldname})")
                lines.append(self._stringify_entries(value, indent + 1))
            else:
                lines.append(f"{indent_str}{field} ({fieldname}): {str(value)}")

        return "\n".join(lines)

    def _check_version(self, value: Value, set_version: bool = False):
        if not isinstance(value, str):
            raise DataTypeError(value, str, SeratoBinFile.Fields.VERSION)
        if value not in self.TESTED_VERSIONS:
            raise ValueError(f"""
                ERROR: Untested version Serato bin file version: {value}
                Please contact the developer so we can get it tested and supported!
                We will have you send the file you are trying to parse so we can add support and tests for it.
                We do not want to risk damaging users' library database or crate files!

                Possible versions:
                {'\n'.join(self.TESTED_VERSIONS)}
                """)
        if set_version:
            self.version = value

    class EntryJson(TypedDict):
        field: "SeratoBinFile.ParsedField"
        fieldname: str
        type: str
        value: str | int | bool | list["SeratoBinFile.EntryJson"]

    def to_json_object(self) -> list[EntryJson]:
        def entries_to_json(entries: SeratoBinFile.EntryList) -> list[SeratoBinFile.EntryJson]:
            result = []

            for field, value in entries:
                if isinstance(value, list):
                    value = entries_to_json(value)
                elif isinstance(value, (bytes, bytearray, memoryview)):
                    value = base64.b64encode(value).decode("utf-8")  # need to make JSON compatible

                entry_obj: SeratoBinFile.EntryJson = {
                    "field": field,
                    "fieldname": SeratoBinFile.get_field_name(field),
                    "value": value,
                    "type": type(value).__name__,
                }

                result.append(entry_obj)

            return result

        return entries_to_json(self.entries)

    def from_json_object(self, json_data: list[EntryJson]):
        def json_to_entries(json_list: list[SeratoBinFile.EntryJson]) -> SeratoBinFile.EntryList:
            result: SeratoBinFile.EntryList = []
            version_set = False
            for entry_obj in json_list:
                field = entry_obj["field"]
                value = entry_obj["value"]
                value_type = entry_obj["type"]

                if value_type == "list":
                    value = json_to_entries(cast(list[SeratoBinFile.EntryJson], value))
                elif value_type in ("bytes", "bytearray", "memoryview"):
                    value = base64.b64decode(cast(bytes, value))
                elif value_type == "int":
                    value = int(cast(int, value))
                elif value_type == "bool":
                    value = bool(cast(bool, value))
                elif value_type == "str":
                    value = str(cast(str, value))
                else:
                    raise ValueError(f"unexpected type: {value_type}")

                if not version_set and field == SeratoBinFile.Fields.VERSION:
                    self._check_version(value, set_version=True)
                    version_set = True

                result.append((field, value))

            return result

        self.entries = json_to_entries(json_data)
        self._dump()

    def write_json(self, filepath: str) -> None:
        def write_json_array(json_entries: list[SeratoBinFile.EntryJson], indent: int = 0) -> str:
            indent_str = "  " * indent
            lines = ["[\n"]

            for i, entry in enumerate(json_entries):
                if isinstance(entry["value"], list):
                    lines.append(
                        f'{indent_str}  {{"field": "{entry["field"]}", "fieldname": "{entry["fieldname"]}", "type": "{entry["type"]}", "value": '
                    )
                    lines.append(write_json_array(entry["value"], indent + 1))
                    lines.append("}")
                else:
                    lines.append(f"{indent_str}  {json.dumps(entry, ensure_ascii=False)}")

                lines.append(",\n" if i < len(json_entries) - 1 else "\n")

            lines.append(f"{indent_str}]")
            return "".join(lines)

        json_content = write_json_array(self.to_json_object(), indent=0)

        with open(filepath, "w", encoding="utf-8") as f:
            json.loads(json_content)  # check
            f.write(json_content)

    def print(self):
        print(self)

    def __repr__(self) -> str:
        return str(self.raw_data)

    class EntryListCls:
        def __init__(self, entries: "SeratoBinFile.EntryList"):
            self.fields: list[str] = []

            for field, value in entries:
                if isinstance(value, list):
                    raise DeeplyNestedListError
                setattr(self, field, value)
                self.fields.append(field)

        def __repr__(self) -> str:
            return str(self.to_entries())

        def get_value(self, field: str) -> "SeratoBinFile.Value":
            return getattr(self, field)

        def set_value(self, field: str, value: "SeratoBinFile.Value"):
            if field not in self.fields:
                self.fields.append(field)
            setattr(self, field, value)

        def to_entries(self) -> "SeratoBinFile.EntryList":
            return [(f, self.get_value(f)) for f in self.fields]

    class Track(EntryListCls):
        def __init__(self, entries: "SeratoBinFile.EntryList", path_key: str):
            super().__init__(entries)

            self.path_key = path_key
            relative_path = self.get_value(path_key)
            if not isinstance(relative_path, str):
                raise DataTypeError(relative_path, str, path_key)
            self.relpath: str = relative_path

        def set_path(self, path: str):
            relpath = SeratoBinFile.get_relative_path(path)
            self.set_value(self.path_key, relpath)
            self.relpath = relpath

        def get_full_path(self):
            return SeratoBinFile.get_full_path(self.relpath)

    def _get_track(self, entries: "SeratoBinFile.EntryList"):
        return SeratoBinFile.Track(entries, path_key=self.TRACK_PATH_KEY)

    @staticmethod
    def _get_type(field: str) -> str:
        # vrsn field has no type_id, but contains text ("t")
        return "t" if field == SeratoBinFile.Fields.VERSION else field[0]

    def _parse_item(self, item_data: bytes) -> Generator["SeratoBinFile.Entry", None, None]:
        fp = io.BytesIO(item_data)
        version_set = False
        for header in iter(lambda: fp.read(8), b""):
            assert len(header) == 8
            field_ascii: bytes
            length: int
            field_ascii, length = struct.unpack(">4sI", header)
            field: str = field_ascii.decode("ascii")
            type_id: str = SeratoBinFile._get_type(field)

            data = fp.read(length)
            assert len(data) == length

            value: SeratoBinFile.Value
            if type_id in ("o", "r"):  #  struct
                value = list(self._parse_item(data))
            elif type_id in ("p", "t"):  # text
                # value = (data[1:] + b"\00").decode("utf-16") # from imported code
                value = data.decode("utf-16-be")
            elif type_id == "b":  # single byte, is a boolean
                value = cast(bool, struct.unpack("?", data)[0])
            elif type_id == "s":  # signed int
                value = cast(int, struct.unpack(">H", data)[0])
            elif type_id == "u":  # unsigned int
                value = cast(int, struct.unpack(">I", data)[0])
            else:
                raise ValueError(f"unexpected type for field: {field}")

            if not version_set and field == SeratoBinFile.Fields.VERSION:
                self._check_version(value, set_version=True)
                version_set = True

            yield field, value

    def _dump_item(self, entry: Entry) -> bytes:
        field, value = entry
        field_bytes = field.encode("ascii")
        assert len(field_bytes) == 4

        if field == SeratoBinFile.Fields.VERSION:
            self._check_version(value)
            if value != self.version:
                raise ValueError(f"version field value {value} does not match parsed value {self.version}")

        type_id: str = SeratoBinFile._get_type(field)

        if type_id in ("o", "r"):  #  struct (list)
            if not isinstance(value, list):
                raise DataTypeError(value, list, field)
            data = self._dump_entries(value)
        elif type_id in ("p", "t"):  # text
            if not isinstance(value, str):
                raise DataTypeError(value, str, field)
            data = value.encode("utf-16-be")
        elif type_id == "b":  # single byte, is a boolean
            if not isinstance(value, bool):
                raise DataTypeError(value, bool, field)
            data = struct.pack("?", value)
        elif type_id == "s":  # signed int
            if not isinstance(value, int):
                raise DataTypeError(value, int, field)
            data = struct.pack(">H", value)
        elif type_id == "u":  # unsigned int
            if not isinstance(value, int):
                raise DataTypeError(value, int, field)
            data = struct.pack(">I", value)
        else:
            raise ValueError(f"unexpected type for field: {field}")

        length = len(data)
        header = struct.pack(">4sI", field_bytes, length)
        return header + data

    def _dump_entries(self, entries: EntryList):
        return b"".join(self._dump_item(entry) for entry in entries)

    def _dump(self):
        self.raw_data = self._dump_entries(self.entries)

    def get_tracks(self) -> Generator[Track, None, None]:
        for field, value in self.entries:
            if field == SeratoBinFile.Fields.TRACK:
                if not isinstance(value, list):
                    raise DataTypeError(value, list, field)
                track = self._get_track(value)
                yield track

    def get_track_paths(self, include_drive: bool = False) -> list[str]:
        return [t.get_full_path() if include_drive else t.relpath for t in self.get_tracks()]

    def modify_tracks(self, func: Callable[[Track], Track]):
        for i, (field, value) in enumerate(self.entries):
            if field == SeratoBinFile.Fields.TRACK:
                if not isinstance(value, list):
                    raise DataTypeError(value, list, field)
                track = self._get_track(value)
                new_track = func(track)
                self.entries[i] = (field, new_track.to_entries())
        self._dump()

    def filter_tracks(self, func: Callable[[Track], bool]):
        new_entries: "SeratoBinFile.EntryList" = []
        for field, value in self.entries:
            if field == SeratoBinFile.Fields.TRACK:
                if not isinstance(value, list):
                    raise DataTypeError(value, list, field)
                track = self._get_track(value)
                if not func(track):
                    continue
            new_entries.append((field, value))
        self.entries = new_entries
        self._dump()

    def remove_track(self, filepath: str):
        self.filter_tracks(lambda track: track.relpath != SeratoBinFile.get_relative_path(filepath))

    def remove_duplicates(self):
        track_paths: list[str] = []

        def filter_track(track: "SeratoBinFile.Track") -> bool:
            was_in_track_paths = track.relpath not in track_paths
            track_paths.append(track.relpath)
            return was_in_track_paths

        self.filter_tracks(filter_track)

    def save(self, file: Optional[str] = None):
        if file is None:
            file = self.filepath
        if file.lower().endswith(".json"):
            raise ValueError("cannot save raw data to .json")
        with open(file, "wb") as f:
            f.write(self.raw_data)

    @staticmethod
    def get_field_name(field: str) -> str:
        try:
            return (
                get_enum_key_from_value(field, SeratoBinFile.Fields)
                .replace("_", " ")
                .title()
                .replace("Smartcrate", "SmartCrate")
                .replace("Added U", "Added")
                .replace("Added T", "Added")
            )
        except ValueError:
            return "Unknown Field"

    @staticmethod
    def _check_valid_field(field: str):
        if field not in SeratoBinFile.FIELDS:
            raise ValueError(
                f"invalid field: {field} must be one of: {str(SeratoBinFile.FIELDS)}\n(see {__file__} for what these keys map to)"
            )

    @staticmethod
    def get_relative_path(filepath: str) -> str:
        filepath = os.path.splitdrive(filepath)[1]
        relpath = os.path.normpath(filepath).replace(os.path.sep, "/").lstrip("/")
        return relpath

    @staticmethod
    def get_full_path(filepath: str):
        drive, filepath = os.path.splitdrive(filepath)
        if drive:
            return filepath
        else:
            return os.path.normpath(os.path.join(SERATO_DRIVE, "\\", filepath))

    def get_entries(self) -> Generator[EntryFull, None, None]:
        """Get entry fieldnames."""
        yield from self.__get_entries_internal(track_matcher=None)

    def get_entries_filtered(self, track_matcher: str | Pattern[str]) -> Generator[EntryFull, None, None]:
        """Get entry fieldnames with track matching filter."""
        yield from self.__get_entries_internal(track_matcher=track_matcher)

    def __get_entries_internal(self, track_matcher: Optional[str | Pattern[str]]) -> Generator[EntryFull, None, None]:
        """Internal implementation for getting entry fieldnames."""
        for field, value in self.entries:
            if isinstance(value, list):
                if track_matcher and field == SeratoBinFile.Fields.TRACK:
                    if not isinstance(value, list):
                        raise DataTypeError(value, list, field)
                    track = self.__class__.Track(value, path_key=self.TRACK_PATH_KEY)
                    if not bool(re.search(track_matcher, track.relpath, re.IGNORECASE)):
                        continue

                try:
                    new_entries: list[SeratoBinFile.EntryFull] = []
                    for f, v in value:
                        if isinstance(v, list):
                            raise DeeplyNestedListError
                        new_entries.append((f, SeratoBinFile.get_field_name(f), v))
                except:
                    logger.error(f"error on field: {field} value: {value}")
                    raise
                value = new_entries

            yield field, SeratoBinFile.get_field_name(field), value

    class ModifyRule(TypedDict):
        field: "SeratoBinFile.Fields"
        func: Callable[[str, "SeratoBinFile.ValueOrNone"], "SeratoBinFile.ValueOrNone"]
        """ (filename: str, prev_value: ValueType | None) -> new_value: ValueType | None """
        files: NotRequired[list[str]]

    def modify(self, rules: list[ModifyRule]):
        # check rule fields
        all_field_names = [rule["field"] for rule in rules]
        assert len(list(rules)) == len(
            list(set(all_field_names))
        ), f"must only have 1 function per field. fields passed: {str(sorted(all_field_names))}"
        for field in all_field_names:
            SeratoBinFile._check_valid_field(field)

        # fix some rules if needed
        for rule in rules:
            if "files" in rule:
                rule["files"] = [SeratoBinFile.get_relative_path(file).upper() for file in rule["files"]]

        def _maybe_perform_rule(field: str, prev_val: "SeratoBinFile.ValueOrNone", track_relpath: str):
            rule = next((r for r in rules if field == r["field"]), None)
            if rule is None:
                return None
            if "files" in rule and track_relpath.upper() not in rule["files"]:
                return None

            maybe_new_value = rule["func"](track_relpath, prev_val)
            if maybe_new_value is None or maybe_new_value == prev_val:
                return None

            if field == self.TRACK_PATH_KEY:
                if not isinstance(maybe_new_value, str):
                    raise DataTypeError(maybe_new_value, str, field)
                if not os.path.exists(SeratoBinFile.get_full_path(maybe_new_value)):
                    raise FileNotFoundError(f"set track location to {maybe_new_value}, but doesn't exist")
                maybe_new_value = SeratoBinFile.get_relative_path(maybe_new_value)

            field_name = SeratoBinFile.get_field_name(field)
            logger.info(f"Set {field}({field_name})={str(maybe_new_value)} in library for {track_relpath}")
            return maybe_new_value

        def modify_track(track: SeratoBinFile.Track) -> SeratoBinFile.Track:
            for f, v in track.to_entries():
                maybe_new_value = _maybe_perform_rule(f, v, track.relpath)
                if maybe_new_value is not None:
                    track.set_value(f, maybe_new_value)
            for rule in rules:
                if rule["field"] not in track.fields:
                    maybe_new_value = _maybe_perform_rule(rule["field"], None, track.relpath)
                    if maybe_new_value is not None:
                        track.set_value(rule["field"], maybe_new_value)
            return track

        self.modify_tracks(modify_track)

    def modify_and_save(self, rules: list[ModifyRule], file: Optional[str] = None):
        self.modify(rules)
        self.save(file)

    def change_track_path(self, src: str, dest: str):
        self.modify([{"field": self.TRACK_PATH_KEY, "files": [src], "func": lambda *args: dest}])

    def find_missing(self):
        new_entries: SeratoBinFile.EntryList = []
        new_dir: str | None = None
        for field, value in self.entries:
            if field == SeratoBinFile.Fields.TRACK:
                if not isinstance(value, list):
                    raise DataTypeError(value, list, field)
                track = self._get_track(value)
                track_path = track.get_full_path()
                if not os.path.isfile(track_path):
                    print(f"missing: {track_path}")
                    new_location = None
                    if new_dir is not None:
                        possible_new_loc = os.path.join(new_dir, os.path.basename(track_path))
                        if os.path.isfile(possible_new_loc):
                            new_location = possible_new_loc
                    if not new_location:
                        while True:
                            new_location = os.path.normpath(
                                input(
                                    'enter new location of file, or directory to look for missing files, or "s" to skip:'
                                )
                                .strip()
                                .strip('"')
                            )
                            if new_location == "s":
                                new_location = None
                                break
                            if os.path.isdir(new_location):
                                new_dir = new_location
                                new_location = os.path.join(new_dir, os.path.basename(track_path))
                            if os.path.exists(new_location):
                                break
                    if new_location:
                        print("   new_location: " + new_location)
                        track.set_path(new_location)
                value = track.to_entries()
            new_entries.append((field, value))
        self.entries = new_entries
        self._dump()
        self.save()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ast
import base64
import configparser
import io
import os
import struct
import sys
from typing import Callable, TypedDict, Literal, List, Sequence, Union, NotRequired
from enum import Enum

from mutagen.mp3 import HeaderNotFoundError

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.track_cues_v1 import TrackCuesV1
from serato_tools.utils import get_enum_key_from_value, logger, DataTypeError
from serato_tools.utils.track_tags import SeratoTag


class TrackCuesV2(SeratoTag):
    GEOB_KEY = "Serato Markers2"
    VERSION = (0x01, 0x01)

    class CueColors(Enum):
        RED = b"\xcc\x00\x00"
        ORANGE = b"\xcc\x44\x00"
        YELLOWORANGE = b"\xcc\x88\x00"
        YELLOW = b"\xcc\xcc\x00"
        LIMEGREEN1 = b"\x88\xcc\x00"
        DARKGREEN = b"\x44\xcc\x00"
        LIMEGREEN2 = b"\x00\xcc\x00"
        LIMEGREEN3 = b"\x00\xcc\x88"
        SEAFOAM = b"\x00\xcc\x88"
        CYAN = b"\x00\xcc\xcc"
        LIGHTBLUE = b"\x00\x88\xcc"
        BLUE1 = b"\x00\x44\xcc"
        BLUE2 = b"\x00\x00\xcc"
        PURPLE1 = b"\x44\x00\xcc"
        PURPLE2 = b"\x88\x00\xcc"
        PINK = b"\xcc\x00\xcc"
        MAGENTA = b"\xcc\x00\x88"
        PINKRED = b"\xcc\x00\x44"

    class TrackColors(Enum):
        PINK = b"\xff\x99\xff"
        DARKPINK = b"\xff\x99\xdd"
        PINKRED = b"\xff\x99\xbb"
        RED = b"\xff\x99\x99"
        ORANGE = b"\xff\xbb\x99"
        YELLOWORANGE = b"\xff\xdd\x99"
        YELLOW = b"\xff\xff\x99"
        LIMEGREEN1 = b"\xdd\xff\x99"
        LIMEGREEN2 = b"\xbb\xff\x99"
        LIMEGREEN3 = b"\x99\xff\x99"
        LIMEGREEN4 = b"\x99\xff\xbb"
        SEAFOAM = b"\x99\xff\xdd"
        CYAN = b"\x99\xff\xff"
        LIGHTBLUE = b"\x99\xdd\xff"
        BLUE1 = b"\x99\xbb\xff"
        BLUE2 = b"\x99\x99\xff"
        PURPLE = b"\xbb\x99\xff"
        MAGENTA = b"\xdd\x99\xff"
        WHITE = b"\xff\xff\xff"
        GREY = b"\xbb\xbb\xbb"
        BLACK = b"\x99\x99\x99"

    def __init__(self, file_or_data: SeratoTag.FileOrData):
        self.raw_data: bytes | None = None
        super().__init__(file_or_data)

        self.entries: list[TrackCuesV2.Entry] = []
        if self.raw_data is not None:
            self.entries = list(self._parse(self.raw_data))
        self.modified: bool = False

    @staticmethod
    def _get_cue_color_key(value: bytes) -> str:
        return get_enum_key_from_value(value, TrackCuesV2.CueColors)

    @staticmethod
    def _get_track_color_key(value: bytes) -> str:
        return get_enum_key_from_value(value, TrackCuesV2.TrackColors)

    @staticmethod
    def _get_entry_class(entry_name: str):
        return next(
            (
                cls
                for cls in (
                    TrackCuesV2.BpmLockEntry,
                    TrackCuesV2.ColorEntry,
                    TrackCuesV2.CueEntry,
                    TrackCuesV2.LoopEntry,
                    TrackCuesV2.FlipEntry,
                )
                if cls.NAME == entry_name
            ),
            TrackCuesV2.UnknownEntry,
        )

    class Entry(object):
        NAME: str | None
        FIELDS: tuple[str, ...]
        data: bytes

        def __init__(self, *args):
            assert len(args) == len(self.FIELDS)
            for field, value in zip(self.FIELDS, args):
                setattr(self, field, value)

        def __repr__(self) -> str:
            return "{name}({data})".format(
                name=self.__class__.__name__,
                data=", ".join("{}={!r}".format(name, getattr(self, name)) for name in self.FIELDS),
            )

        @classmethod
        def load(cls, data: bytes):
            return cls(data)

        def dump(self) -> bytes:
            return self.data

    class UnknownEntry(Entry):
        NAME = None
        FIELDS = ("data",)

        @classmethod
        def load(cls, data: bytes):
            return cls(data)

        def dump(self):
            return self.data

    class BpmLockEntry(Entry):
        NAME = "BPMLOCK"
        FIELDS = ("enabled",)
        FORMAT = "?"

        @classmethod
        def load(cls, data: bytes):
            return cls(*struct.unpack(cls.FORMAT, data))

        def dump(self):
            return struct.pack(self.FORMAT, *(getattr(self, f) for f in self.FIELDS))

    class ColorEntry(Entry):
        NAME = "COLOR"
        FORMAT = "c3s"
        FIELDS = (
            "field1",
            "color",
        )

        @classmethod
        def load(cls, data: bytes):
            return cls(*struct.unpack(cls.FORMAT, data))

        def dump(self):
            return struct.pack(self.FORMAT, *(getattr(self, f) for f in self.FIELDS))

    class CueEntry(Entry):
        NAME = "CUE"
        FORMAT = ">cBIc3s2s"
        FIELDS = (
            "field1",
            "index",
            "position",
            "field4",
            "color",
            "field6",
            "name",
        )
        name: str

        @classmethod
        def load(cls, data: bytes):
            info_size = struct.calcsize(cls.FORMAT)
            info = struct.unpack(cls.FORMAT, data[:info_size])
            name, nullbyte, other = data[info_size:].partition(b"\x00")
            assert nullbyte == b"\x00"
            assert other == b""
            return cls(*info, name.decode("utf-8"))

        def dump(self):
            struct_fields = self.FIELDS[:-1]
            return b"".join(
                (
                    struct.pack(self.FORMAT, *(getattr(self, f) for f in struct_fields)),
                    self.name.encode("utf-8"),
                    b"\x00",
                )
            )

    class LoopEntry(Entry):
        NAME = "LOOP"
        FORMAT = ">cBII4s4sB?"
        FIELDS = (
            "field1",
            "index",
            "startposition",
            "endposition",
            "field5",
            "field6",
            "color",
            "locked",
            "name",
        )
        name: str

        @classmethod
        def load(cls, data: bytes):
            info_size = struct.calcsize(cls.FORMAT)
            info = struct.unpack(cls.FORMAT, data[:info_size])
            name, nullbyte, other = data[info_size:].partition(b"\x00")
            assert nullbyte == b"\x00"
            assert other == b""
            return cls(*info, name.decode("utf-8"))

        def dump(self):
            struct_fields = self.FIELDS[:-1]
            return b"".join(
                (
                    struct.pack(self.FORMAT, *(getattr(self, f) for f in struct_fields)),
                    self.name.encode("utf-8"),
                    b"\x00",
                )
            )

    class FlipEntry(Entry):
        NAME = "FLIP"
        FORMAT1 = "cB?"
        FORMAT2 = ">BI"
        FORMAT3 = ">BI16s"
        FIELDS = (
            "field1",
            "index",
            "enabled",
            "name",
            "loop",
            "num_actions",
            "actions",
        )

        @classmethod
        def load(cls, data):
            info1_size = struct.calcsize(cls.FORMAT1)
            info1 = struct.unpack(cls.FORMAT1, data[:info1_size])
            name, nullbyte, other = data[info1_size:].partition(b"\x00")
            assert nullbyte == b"\x00"

            info2_size = struct.calcsize(cls.FORMAT2)
            loop, num_actions = struct.unpack(cls.FORMAT2, other[:info2_size])
            action_data = other[info2_size:]
            actions = []
            for i in range(num_actions):  # pylint: disable=unused-variable
                type_id, size = struct.unpack(cls.FORMAT2, action_data[:info2_size])
                action_data = action_data[info2_size:]
                if type_id == 0:
                    payload = struct.unpack(">dd", action_data[:size])
                    actions.append(("JUMP", *payload))
                elif type_id == 1:
                    payload = struct.unpack(">ddd", action_data[:size])
                    actions.append(("CENSOR", *payload))
                action_data = action_data[size:]
            assert action_data == b""

            return cls(*info1, name.decode("utf-8"), loop, num_actions, actions)

        def dump(self):
            raise NotImplementedError("FLIP entry dumps are not implemented!")

    def _parse(self, data: bytes):
        self._check_version(data[: self.VERSION_LEN])
        data = data[self.VERSION_LEN :]  # b64data
        try:
            end_index = data.index(b"\x00")
            data = data[:end_index]
        except ValueError:
            pass  # data = data
        data = data.replace(b"\n", b"")
        padding = b"A==" if len(data) % 4 == 1 else (b"=" * (-len(data) % 4))
        payload = base64.b64decode(data + padding)
        fp = io.BytesIO(payload)
        self._check_version(fp.read(self.VERSION_LEN))
        while True:
            entry_name = SeratoTag._readbytes(fp).decode("utf-8")
            if not entry_name:
                break
            entry_len = struct.unpack(">I", fp.read(4))[0]
            assert entry_len > 0

            entry_class = TrackCuesV2._get_entry_class(entry_name)
            yield entry_class.load(fp.read(entry_len))

    def _dump(self):
        version = self._pack_version()

        contents: list[bytes] = [version]
        for entry in self.entries:
            data = entry.dump()
            if entry.NAME is not None:
                data = b"".join(
                    (
                        entry.NAME.encode("utf-8"),
                        b"\x00",
                        struct.pack(">I", (len(data))),
                        data,
                    )
                )
            contents.append(data)

        payload = b"".join(contents)
        payload_base64 = bytearray(base64.b64encode(payload).replace(b"=", b"A"))

        i = 72
        while i < len(payload_base64):
            payload_base64.insert(i, 0x0A)
            i += 73

        data = version
        data += payload_base64

        new_raw_data = data.ljust(470, b"\x00")

        self.modified = new_raw_data != self.raw_data  # pylint: disable=access-member-before-definition
        self.raw_data = new_raw_data

    @staticmethod
    def parse_entries_file(contents: str, assert_len_1: bool):
        cp = configparser.ConfigParser()
        cp.read_string(contents)
        sections: Sequence[str] = tuple(sorted(cp.sections()))
        if assert_len_1:
            assert len(sections) == 1

        results: list[TrackCuesV2.Entry] = []
        for section in sections:
            l, s, r = section.partition(": ")
            entry_class = TrackCuesV2._get_entry_class(r if s else l)

            e = entry_class(
                *(
                    ast.literal_eval(
                        cp.get(section, field),
                    )
                    for field in entry_class.FIELDS
                )
            )
            results.append(entry_class.load(e.dump()))
        return results

    type Value = bytes | str | int

    class EntryModifyRule(TypedDict):
        field: str
        func: Callable[["TrackCuesV2.Value"], Union["TrackCuesV2.Value", None]]
        """ (prev_value: ValueType) -> new_value: ValueType | None """

    class CueIndexModifyRule(EntryModifyRule):
        field: Literal["index"]  # pyright: ignore[reportIncompatibleVariableOverride]
        func: Callable[[int], int | None]
        """ (prev_value: ValueType) -> new_value: ValueType | None """

    class CueColorModifyRule(EntryModifyRule):
        field: Literal["color"]  # pyright: ignore[reportIncompatibleVariableOverride]
        func: Callable[["TrackCuesV2.CueColors"], Union["TrackCuesV2.CueColors", None]]
        """ (prev_value: ValueType) -> new_value: ValueType | None """

    class CueNameModifyRule(EntryModifyRule):
        field: Literal["name"]  # pyright: ignore[reportIncompatibleVariableOverride]
        func: Callable[[str], str | None]
        """ (prev_value: ValueType) -> new_value: ValueType | None """

    class TrackColorModifyRule(EntryModifyRule):
        field: Literal["color"]  # pyright: ignore[reportIncompatibleVariableOverride]
        func: Callable[["TrackCuesV2.TrackColors"], Union["TrackCuesV2.TrackColors", None]]
        """ (prev_value: ValueType) -> new_value: ValueType | None """

    type CueRules = CueIndexModifyRule | CueColorModifyRule | CueNameModifyRule | EntryModifyRule
    type TrackColorRules = TrackColorModifyRule | EntryModifyRule

    class EntryModifyRules(TypedDict):
        cues: NotRequired[List["TrackCuesV2.CueRules"]]
        color: NotRequired[List["TrackCuesV2.TrackColorRules"]]

    FIELD_TO_TYPE_MAP = {"color": bytes, "index": int, "name": str}

    @staticmethod
    def _modify_entry(entry: Entry, rules: Sequence[CueRules | TrackColorRules]):
        """
        Returns:
            entry: entry if was modified. If was not changed, returns None.
        """

        all_field_names = [rule["field"] for rule in rules]
        assert len(rules) == len(
            list(set(all_field_names))
        ), f"must only have 1 function per field. fields passed: {str(sorted(all_field_names))}"
        # TODO: ensure field is valid else throw error!

        change_made = False

        output = f"[{entry.NAME}]\n"
        for field in entry.FIELDS:
            value: TrackCuesV2.Value = getattr(entry, field)

            rule = next((r for r in rules if field == r["field"]), None)
            if rule:
                ExpectedType: type | None = TrackCuesV2.FIELD_TO_TYPE_MAP.get(field, None)
                if ExpectedType and not isinstance(value, ExpectedType):
                    raise DataTypeError(value, ExpectedType, field)

                prev_value = value
                if field == "color":
                    is_track_color = isinstance(entry, TrackCuesV2.ColorEntry)
                    Colors = TrackCuesV2.TrackColors if is_track_color else TrackCuesV2.CueColors
                    try:
                        prev_value = Colors(prev_value)
                    except ValueError:
                        pass

                maybe_new_val = rule["func"](prev_value)  # type: ignore
                if isinstance(maybe_new_val, Enum):
                    maybe_new_val = maybe_new_val.value
                if maybe_new_val is not None and maybe_new_val != value:
                    value = maybe_new_val

                    if ExpectedType and not isinstance(value, ExpectedType):
                        raise DataTypeError(value, ExpectedType, field)

                    change_made = True

                    if field == "color":
                        is_track_color = isinstance(entry, TrackCuesV2.ColorEntry)
                        color_name: str | None = None
                        if isinstance(value, bytes):
                            get_color_name = (
                                TrackCuesV2._get_track_color_key if is_track_color else TrackCuesV2._get_cue_color_key
                            )
                            try:
                                color_name = get_color_name(value)
                            except ValueError:
                                color_name = None
                        color_log_str = "Track" if is_track_color else "Cue"
                        logger.info(
                            f"Set {color_log_str} Color to {color_name if color_name else f'Unknown Color ({str(value)})'}"
                        )
                    else:
                        logger.info(f'Set {type(entry).__name__} field "{field}" to {str(value)}')
            output += f"{field}: {value!r}\n"
        output += "\n"

        if not change_made:
            return None

        entry = TrackCuesV2.parse_entries_file(output, assert_len_1=True)[0]
        return entry

    def modify_entries(self, rules: EntryModifyRules, delete_tags_v1: bool = True):
        """
        Args:
            delete_tags_v1: Must delete delete_tags_v1 in order for many tags_v2 changes appear in Serato (since we never change tags_v1 along with it (TODO)). Not sure what tags_v1 is even for, probably older versions of Serato. Have found no issues with deleting this, but use with caution if running an older version of Serato.
        """
        if delete_tags_v1 and self.tagfile:
            super(SeratoTag, self)._del_geob(TrackCuesV1.GEOB_KEY)  # pylint: disable=bad-super-call

        new_entries = []
        change_made = False
        for entry in self.entries:
            maybe_new_entry = None
            if "cues" in rules and isinstance(entry, TrackCuesV2.CueEntry):
                maybe_new_entry = TrackCuesV2._modify_entry(entry, rules["cues"])
            elif "color" in rules and isinstance(entry, TrackCuesV2.ColorEntry):
                maybe_new_entry = TrackCuesV2._modify_entry(entry, rules["color"])
            if maybe_new_entry is not None:
                entry = maybe_new_entry
                change_made = True
            new_entries.append(entry)

        if not change_made:
            return

        self.entries = new_entries
        self._dump()

    def save(self, force: bool = False):
        if self.modified or force:
            super().save()

    def get_track_color(self) -> bytes | None:
        color_entry = next(
            (entry for entry in self.entries if isinstance(entry, TrackCuesV2.ColorEntry)),
            None,
        )
        if color_entry is None:
            return None
        return getattr(color_entry, "color")

    def get_track_color_name(self) -> str | None:
        color_bytes = self.get_track_color()
        if color_bytes is None:
            return None
        return self._get_track_color_key(color_bytes)

    def set_track_color(self, color: TrackColors, delete_tags_v1: bool = True):
        """
        Args:
            color: TrackColors (bytes)
            delete_tags_v1: Must delete delete_tags_v1 in order for track color change to appear in Serato (since we never change tags_v1 along with it (TODO)). Not sure what tags_v1 is even for, probably older versions of Serato. Have found no issues with deleting this, but use with caution if running an older version of Serato.
        """
        self.modify_entries({"color": [{"field": "color", "func": lambda v: color}]}, delete_tags_v1)

    def is_beatgrid_locked(self) -> bool:
        return any(
            (isinstance(entry, TrackCuesV2.BpmLockEntry) and getattr(entry, "enabled")) for entry in self.entries
        )


if __name__ == "__main__":
    import argparse
    import math
    import subprocess
    import tempfile

    from serato_tools.utils.ui import get_hex_editor, get_text_editor, ui_ask

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument(
        "--set_color",
        dest="set_color",
        default=None,
        help="Set track color",
    )
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    try:
        tags = TrackCuesV2(args.file)
    except HeaderNotFoundError:
        with open(args.file, mode="rb") as fp:
            data = fp.read()
        tags = TrackCuesV2(data)

    if args.set_color:
        tags.set_track_color(args.set_color)
        tags.save()
        sys.exit()

    if args.edit:
        text_editor = get_text_editor()
        hex_editor = get_hex_editor()

    new_entries: list[TrackCuesV2.Entry] = []
    width = math.floor(math.log10(len(tags.entries))) + 1
    action = None
    for entry_index, entry in enumerate(tags.entries):
        if args.edit:
            if action not in ("q", "_"):
                print("{:{}d}: {!r}".format(entry_index, width, entry))
                action = ui_ask(
                    "Edit this entry",
                    {
                        "y": "edit this entry",
                        "n": "do not edit this entry",
                        "q": ("quit; do not edit this entry or any of the " "remaining ones"),
                        "a": "edit this entry and all later entries in the file",
                        "b": "edit raw bytes",
                        "r": "remove this entry",
                    },
                    default="n",
                )

            if action in ("y", "a", "b"):
                while True:
                    with tempfile.NamedTemporaryFile() as f:
                        if action == "b":
                            f.write(entry.dump())
                            editor = hex_editor  # pyright: ignore[reportPossiblyUnboundVariable]
                        else:
                            if action == "a":
                                entries_to_edit = (
                                    (
                                        "{:{}d}: {}".format(i, width, e.NAME),
                                        e,
                                    )
                                    for i, e in enumerate(tags.entries[entry_index:], start=entry_index)
                                )
                            else:
                                entries_to_edit = ((entry.NAME, entry),)

                            for section, e in entries_to_edit:
                                f.write("[{}]\n".format(section).encode())
                                for field in e.FIELDS:
                                    f.write("{}: {!r}\n".format(field, getattr(e, field)).encode())
                                f.write(b"\n")
                            editor = text_editor  # pyright: ignore[reportPossiblyUnboundVariable]
                        f.flush()
                        status = subprocess.call((editor, f.name))
                        f.seek(0)
                        output = f.read()

                    if status != 0:
                        if (
                            ui_ask(
                                "Command failed, retry",
                                {
                                    "y": "edit again",
                                    "n": "leave unchanged",
                                },
                            )
                            == "n"
                        ):
                            break
                    else:
                        try:
                            if action != "b":
                                results = TrackCuesV2.parse_entries_file(output.decode(), assert_len_1=action != "a")
                            else:
                                results = [entry.load(output)]
                        except Exception as e:  # pylint: disable=broad-exception-caught
                            print(str(e))
                            if (
                                ui_ask(
                                    "Content seems to be invalid, retry",
                                    {
                                        "y": "edit again",
                                        "n": "leave unchanged",
                                    },
                                )
                                == "n"
                            ):
                                break
                        else:
                            for i, e in enumerate(results, start=entry_index):
                                print("{:{}d}: {!r}".format(i, width, e))
                            subaction = ui_ask(
                                "Above content is valid, save changes",
                                {
                                    "y": "save current changes",
                                    "n": "discard changes",
                                    "e": "edit again",
                                },
                                default="y",
                            )
                            if subaction == "y":
                                new_entries.extend(results)
                                if action == "a":
                                    action = "_"
                                break
                            elif subaction == "n":
                                if action == "a":
                                    action = "q"
                                new_entries.append(entry)
                                break
            elif action in ("r", "_"):
                continue
            else:
                new_entries.append(entry)
        else:
            print("{:{}d}: {!r}".format(entry_index, width, entry))

    if args.edit:
        if new_entries == tags.entries:
            print("No changes made.")
        else:
            tags.entries = new_entries
            tags._dump()  # pylint: disable=protected-access
            if tags.tagfile:
                tags.save()
            else:
                with open(args.file, mode="wb") as fp:
                    fp.write(tags.raw_data or b"")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import collections
import io
import struct
import os
import sys
from typing import cast

from mutagen.mp3 import HeaderNotFoundError

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.track_tags import SeratoTag
from serato_tools.utils import logger


class TrackBeatgrid(SeratoTag):
    GEOB_KEY = "Serato BeatGrid"
    VERSION = (0x01, 0x00)

    NonTerminalBeatgridMarker = collections.namedtuple(
        "NonTerminalBeatgridMarker",
        ("position", "beats_till_next_marker"),
    )
    TerminalBeatgridMarker = collections.namedtuple(
        "TerminalBeatgridMarker",
        ("position", "bpm"),
    )

    Footer = collections.namedtuple("Footer", ("unknown",))

    type EntryList = list[TerminalBeatgridMarker | NonTerminalBeatgridMarker | Footer]

    def __init__(self, file_or_data: SeratoTag.FileOrData):
        self.raw_data: bytes | None = None
        super().__init__(file_or_data)

        self.entries: TrackBeatgrid.EntryList | None = None

        if self.raw_data is not None:
            self.entries = list(self._parse(self.raw_data))

    def __str__(self) -> str:
        nonterminal_markers, terminal_markers, footer = self._check_and_split()  # pylint: disable=unused-variable
        markers = nonterminal_markers + terminal_markers
        return f"Beatgrid with {len(markers)} markers"

    def _parse(self, data: bytes):
        fp = io.BytesIO(data)
        self._check_version(fp.read(self.VERSION_LEN))

        num_markers = struct.unpack(">I", fp.read(4))[0]
        for i in range(num_markers):
            position = struct.unpack(">f", fp.read(4))[0]
            data = fp.read(4)
            if i == num_markers - 1:
                bpm = struct.unpack(">f", data)[0]
                yield TrackBeatgrid.TerminalBeatgridMarker(position, bpm)
            else:
                beats_till_next_marker = struct.unpack(">I", data)[0]
                yield TrackBeatgrid.NonTerminalBeatgridMarker(position, beats_till_next_marker)

        # TODO: What's the meaning of the footer byte?
        yield TrackBeatgrid.Footer(struct.unpack("B", fp.read(1))[0])
        assert fp.read() == b""

    def _check_and_split(self):
        if not self.entries:
            raise ValueError("no entries set")

        nonterminal_markers: list[TrackBeatgrid.NonTerminalBeatgridMarker] = []
        terminal_markers: list[TrackBeatgrid.TerminalBeatgridMarker] = []
        footers: list[TrackBeatgrid.Footer] = []

        for entry in self.entries:
            if isinstance(entry, TrackBeatgrid.NonTerminalBeatgridMarker):
                nonterminal_markers.append(entry)
            elif isinstance(entry, TrackBeatgrid.TerminalBeatgridMarker):
                terminal_markers.append(entry)
            elif isinstance(entry, TrackBeatgrid.Footer):
                footers.append(entry)
            else:
                raise TypeError(f"unexpected value type {entry}")

        assert len(terminal_markers) == 1, f"should only be 1 terminal marker, but #: {len(terminal_markers)}"
        assert len(footers) == 1, f"should only be 1 footer, but #: {len(footers)}"
        assert isinstance(self.entries[-1], TrackBeatgrid.Footer), "last item should be a footer"
        assert isinstance(
            self.entries[-2], TrackBeatgrid.TerminalBeatgridMarker
        ), "last item should be a terminal marker"
        return nonterminal_markers, terminal_markers, footers[0]

    def _dump(self):
        nonterminal_markers, terminal_markers, footer = self._check_and_split()
        markers = nonterminal_markers + terminal_markers

        fp = io.BytesIO()
        # Write version
        fp.write(self._pack_version())

        # Write markers
        fp.write(struct.pack(">I", len(markers)))
        for marker in markers:
            fp.write(struct.pack(">f", marker.position))
            if isinstance(marker, TrackBeatgrid.TerminalBeatgridMarker):
                fp.write(struct.pack(">f", marker.bpm))
            elif isinstance(marker, TrackBeatgrid.NonTerminalBeatgridMarker):
                fp.write(struct.pack(">I", marker.beats_till_next_marker))
            else:
                raise TypeError(f"Unexpected marker type: {type(marker)}")

        # Write footer
        fp.write(struct.pack("B", footer.unknown))

        # Set to self.raw_data
        fp.seek(0)
        self.raw_data = fp.getvalue()

    def analyze_and_write(self):
        if not self.tagfile:
            raise Exception("No tagfile set")

        from serato_tools.utils.beatgrid_analyze import analyze_beatgrid

        bpm = float(str(self.tagfile["TBPM"]))
        filename = cast(str, self.tagfile.filename)

        logger.info("Analyzing beat grid...")
        analyzed_breatgrid = analyze_beatgrid(filename, bpm_helper=bpm)

        logger.info("Writing tags...")
        entries: TrackBeatgrid.EntryList = [
            TrackBeatgrid.NonTerminalBeatgridMarker(position, 4) for position in analyzed_breatgrid.downbeats[:-1]
        ] + [
            TrackBeatgrid.TerminalBeatgridMarker(analyzed_breatgrid.downbeats[-1], bpm=bpm or analyzed_breatgrid.bpm),
            TrackBeatgrid.Footer(0),
        ]

        self.entries = entries

        self._dump()
        self.save()


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-a", "--analyze", action="store_true", help="Analyze dynamic beatgrid and write to file")
    args = parser.parse_args()

    try:
        tags = TrackBeatgrid(args.file)
    except HeaderNotFoundError:
        with open(args.file, mode="rb") as fp:
            data = fp.read()
        tags = TrackBeatgrid(data)

    if args.analyze and tags.tagfile:
        tags.analyze_and_write()
    else:
        if not tags.entries:
            raise ValueError("no entries")
        for entry in tags.entries:
            print(entry)


def main_analyze():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()
    tags = TrackBeatgrid(args.file)
    tags.analyze_and_write()


if __name__ == "__main__":
    main()



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.crate import Crate

with open("test/data/crate.txt", "r", encoding="utf-8") as f:
    base_expected = f.read()


class TestCase(unittest.TestCase):
    def test_format_filepath(self):
        self.assertEqual(
            Crate.get_relative_path("C:\\Music\\DJ Tracks\\Zeds Dead - In The Beginning.mp3"),
            "Music/DJ Tracks/Zeds Dead - In The Beginning.mp3",
        )
        self.assertEqual(
            Crate.get_relative_path("Music\\DJ Tracks\\Zeds Dead - In The Beginning.mp3"),
            "Music/DJ Tracks/Zeds Dead - In The Beginning.mp3",
        )
        self.assertEqual(
            Crate.get_relative_path("C:/Music/DJ Tracks/Tripp St. - Enlighten.mp3"),
            "Music/DJ Tracks/Tripp St. - Enlighten.mp3",
        )
        self.assertEqual(
            Crate.get_relative_path("Music/DJ Tracks/Tripp St. - Enlighten.mp3"),
            "Music/DJ Tracks/Tripp St. - Enlighten.mp3",
        )

    def test_parse_and_modify_and_dump(self):
        file = os.path.abspath("test/data/TestCrate.crate")
        with open(file, mode="rb") as fp:
            file_data = fp.read()

        crate = Crate(file)

        self.maxDiff = None

        self.assertEqual(crate.raw_data, file_data, "raw_data read")
        expected = base_expected

        self.assertEqual(crate.__str__(), expected, "parse")

        crate.add_track("C:\\Users\\bvand\\Music\\DJ Tracks\\Soulacybin - Zeu.mp3")
        expected += "\notrk (Track): [ ptrk (Track Path): Users/bvand/Music/DJ Tracks/Soulacybin - Zeu.mp3 ]"
        self.assertEqual(crate.__str__(), expected, "track added")

        for t in [
            "C:\\Users\\bvand\\Music\\DJ Tracks\\Soulacybin - Zeu.mp3",
            "Users\\bvand\\Music\\DJ Tracks\\Soulacybin - Zeu.mp3",
            "C:/Users/bvand/Music/DJ Tracks/Soulacybin - Zeu.mp3",
            "Users/bvand/Music/DJ Tracks/Soulacybin - Zeu.mp3",
            "/Users/bvand/Music/DJ Tracks/Soulacybin - Zeu.mp3",
        ]:
            crate.add_track(t)
            self.assertEqual(crate.__str__(), expected, "duplicate track not added")

        crate.add_track("C:/Users/bvand/Music/DJ Tracks/Thundercat - Them Changes.mp3")
        expected += "\notrk (Track): [ ptrk (Track Path): Users/bvand/Music/DJ Tracks/Thundercat - Them Changes.mp3 ]"
        self.assertEqual(crate.__str__(), expected, "track added")



import logging
import os
from typing import Iterable, TypeVar, Type, cast, Any
from enum import Enum

T = TypeVar("T")

logger = logging.getLogger("serato-tools")

SERATO_DIR_NAME = "_Serato_"
SERATO_DIR = os.path.join(os.path.expanduser("~"), "Music", SERATO_DIR_NAME)
SERATO_DRIVE = os.path.splitdrive(SERATO_DIR)[0]


def get_key_from_value(value: T, dict: dict[str, T]) -> str:
    for key, v in dict.items():
        if v == value:
            return key
    raise ValueError(f"no key for value {value}")


def parse_cli_keys_and_values(args: list[str]):
    keys_and_values: dict[str, list[str | int]] = {}
    key = None
    for arg in args:
        if arg.startswith("--"):
            key = arg.lstrip("-")
            keys_and_values[key] = []
        elif key:
            keys_and_values[key].append(arg)

    # pairs[key] = True  # default to True if no value follows
    return keys_and_values


def get_enum_key_from_value(value: str | bytes | int, enum_class: Type[Enum]):
    for member in enum_class:
        if member.value == value:
            return member.name
    raise ValueError(f"no key for value {value}")


def to_array(x: T | Iterable[T]) -> Iterable[T]:
    if isinstance(x, (str, bytes)):
        return cast(list[T], [x])
    if isinstance(x, Iterable):
        return x
    return [x]


class DataTypeError(Exception):
    def __init__(self, value: Any, expected_type: type | Iterable[type], field: str | None):
        super().__init__(
            f"value must be {' or '.join(e.__name__ for e in to_array(expected_type))} when field is {field} (type: {type(value).__name__}) (value: {str(value)})"
        )


class DeeplyNestedListError(Exception):
    def __init__(self):
        super().__init__("unexpected type, deeply nested list. if this occurs, need to implement code for it.")



# pylint: disable=protected-access
import unittest
import os

from src.serato_tools.smart_crate import SmartCrate

with open("test/data/smart_crate.txt", "r", encoding="utf-8") as f:
    expected = f.read()


class TestCase(unittest.TestCase):
    def check_lines(self, given: str, exp: str):
        given = "\n".join([l.strip() for l in given.splitlines()])
        exp = "\n".join([l.strip() for l in exp.splitlines()])
        self.maxDiff = None
        self.assertMultiLineEqual(given, exp)

    def test_parse(self):
        file = os.path.abspath("test/data/TestSmartCrate.scrate")
        with open(file, mode="rb") as fp:
            file_data = fp.read()

        crate = SmartCrate(file)

        self.assertEqual(crate.raw_data, file_data, "raw_data read")
        self.check_lines(crate.__str__(), expected)

    def test_add_rule(self):
        file = os.path.abspath("test/data/TestSmartCrate.scrate")

        crate = SmartCrate(file)
        crate.set_rule(SmartCrate.RuleField.ALBUM, SmartCrate.RuleComparison.STR_IS_NOT, "albo")
        self.check_lines(
            crate.__str__(),
            expected
            + "\nrurt (SmartCrate Rule): [ trft (Rule Comparison): cond_isn_str (STR_IS_NOT) ], [ urkt (Rule Field): 8 (album) ], [ trpt (Rule Value Text): albo ]",
        )



import os
import sys
from typing import Optional

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.track_tags import SeratoTrack


class TrackGain(SeratoTrack):
    REPLAY_GAIN_GAIN_KEY = "replaygain_SeratoGain_gain"
    REPLAY_GAIN_PEAK_KEY = "replaygain_SeratoGain_peak"

    def __init__(self, file: SeratoTrack.FileArg):
        super().__init__(file)
        self.gain: float | None = self.tagfile.get(TrackGain.REPLAY_GAIN_GAIN_KEY, None)
        self.peak: float | None = self.tagfile.get(TrackGain.REPLAY_GAIN_PEAK_KEY, None)

    def __str__(self) -> str:
        return f"gain: {self.gain}\npeak: {self.peak}"

    def set_and_save(self, gain: Optional[float] = None, peak: Optional[float] = None):
        if gain is not None:
            self.gain = gain
        if peak is not None:
            self.peak = peak

        self.tagfile[TrackGain.REPLAY_GAIN_GAIN_KEY] = self.gain
        self.tagfile[TrackGain.REPLAY_GAIN_PEAK_KEY] = self.peak
        self.tagfile.save()

    def save(self):
        self.tagfile.save()

    def delete(self):
        for key in [TrackGain.REPLAY_GAIN_GAIN_KEY, TrackGain.REPLAY_GAIN_PEAK_KEY]:
            if key in self.tagfile:
                del self.tagfile[key]


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()

    tags = TrackGain(args.file)
    print(str(tags))
